# Copyright (c) 2020 The Khronos Group Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import tarfile


def tgz_compress(dir_path, file_path, compression_level=0):
    target_directory = os.path.dirname(file_path)
    if target_directory and not os.path.exists(target_directory):
        os.makedirs(target_directory)

    with tarfile.open(file_path, 'w:gz', compresslevel=compression_level) as tar:
        for file_ in os.listdir(dir_path):
            tar.add(dir_path + '/' + file_, file_)


def tgz_extract(file_path, dir_path):
    if dir_path and not os.path.exists(dir_path):
        os.makedirs(dir_path)

    with tarfile.open(file_path, 'r:gz') as tar:
        tar.extractall(dir_path)



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: tflite

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class Model(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsModel(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Model()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def ModelBufferHasIdentifier(cls, buf, offset, size_prefixed=False):
        return flatbuffers.util.BufferHasIdentifier(buf, offset, b"\x54\x46\x4C\x33", size_prefixed=size_prefixed)

    # Model
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Model
    def Version(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint32Flags, o + self._tab.Pos)
        return 0

    # Model
    def OperatorCodes(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from .OperatorCode import OperatorCode
            obj = OperatorCode()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Model
    def OperatorCodesLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Model
    def OperatorCodesIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        return o == 0

    # Model
    def Subgraphs(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from .SubGraph import SubGraph
            obj = SubGraph()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Model
    def SubgraphsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Model
    def SubgraphsIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        return o == 0

    # Model
    def Description(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Model
    def Buffers(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from .Buffer import Buffer
            obj = Buffer()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Model
    def BuffersLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Model
    def BuffersIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        return o == 0

    # Model
    def MetadataBuffer(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(flatbuffers.number_types.Int32Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return 0

    # Model
    def MetadataBufferAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Int32Flags, o)
        return 0

    # Model
    def MetadataBufferLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Model
    def MetadataBufferIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        return o == 0

    # Model
    def Metadata(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from .Metadata import Metadata
            obj = Metadata()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Model
    def MetadataLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Model
    def MetadataIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        return o == 0

def ModelStart(builder): builder.StartObject(7)
def ModelAddVersion(builder, version): builder.PrependUint32Slot(0, version, 0)
def ModelAddOperatorCodes(builder, operatorCodes): builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(operatorCodes), 0)
def ModelStartOperatorCodesVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def ModelAddSubgraphs(builder, subgraphs): builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(subgraphs), 0)
def ModelStartSubgraphsVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def ModelAddDescription(builder, description): builder.PrependUOffsetTRelativeSlot(3, flatbuffers.number_types.UOffsetTFlags.py_type(description), 0)
def ModelAddBuffers(builder, buffers): builder.PrependUOffsetTRelativeSlot(4, flatbuffers.number_types.UOffsetTFlags.py_type(buffers), 0)
def ModelStartBuffersVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def ModelAddMetadataBuffer(builder, metadataBuffer): builder.PrependUOffsetTRelativeSlot(5, flatbuffers.number_types.UOffsetTFlags.py_type(metadataBuffer), 0)
def ModelStartMetadataBufferVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def ModelAddMetadata(builder, metadata): builder.PrependUOffsetTRelativeSlot(6, flatbuffers.number_types.UOffsetTFlags.py_type(metadata), 0)
def ModelStartMetadataVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def ModelEnd(builder): return builder.EndObject()



# Copyright (c) 2020 The Khronos Group Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .protobuf import *
from .writer import _build_attribute
from .reader import _get_attributes
import numpy as np
import six
try:
    import tensorflow.compat.v1 as tf
except ImportError:
    import tensorflow as tf


def import_graph_def(graph_def):
    graph = tf.Graph()
    with graph.as_default():
        tf.import_graph_def(graph_def, name='')
    return graph


def export_graph_def(graph):
    return graph.as_graph_def(add_shapes=True)


def reinfer_shapes(graph_def):
    return export_graph_def(import_graph_def(graph_def))


def _try_eval(tensor, session):
    try:
        if tensor.dtype == tf.resource or tensor.dtype == tf.string:
            return None
        value = tensor.eval(session=session)
        print("Evaluated constant tensor '{}'".format(tensor.name))
        return value
    except:
        return None


def _build_node(type, name, attribs, inputs):
    node_def = NodeDef()
    node_def.op = type
    node_def.name = name
    if len(inputs):
        node_def.input.extend(inputs)
    for name, value in attribs.items():
        _build_attribute(node_def.attr[name], value)
    return node_def


def _make_const_node(value, name):
    return _build_node('Const', name, {'dtype': value.dtype.type, 'value': value, '_output_shapes': [value.shape]}, [])


def _make_identity_node(input, name, dtype, shape):
    return _build_node('Identity', name, {'T': dtype, '_output_shapes': [shape]}, [input])


def _freeze_shape_tensors(graph_def):
    graph = import_graph_def(graph_def)

    evaluated = {}
    for op in graph.get_operations():
        if op.type == 'Shape':
            shape = op.inputs[0].shape
            if shape.dims is not None and all(item is not None for item in shape.as_list()):
                evaluated[op.name] = np.array(shape, dtype=np.int32)
                print("Evaluated Shape op '{}' to {}".format(op.name, str(shape)))

    changed = False
    new_graph_def = GraphDef()
    for node in graph_def.node:
        value = evaluated.get(node.name)
        if value is not None:
            new_graph_def.node.append(_make_const_node(value, node.name))
            changed = True
        else:
            new_graph_def.node.append(node)

    return new_graph_def, changed


def _remove_const_control_dependencies(graph_def):
    for node in graph_def.node:
        if node.op == 'Const':
            for idx in reversed(range(len(node.input))):
                name = node.input[idx]
                if name[0] == '^':
                    del node.input[idx]

    return graph_def


def _remove_zero_index(name):
    return name[:-2] if name.endswith(':0') else name


def _remove_const_identities(graph_def):
    graph = import_graph_def(graph_def)

    removables = {op.name: _remove_zero_index(op.inputs[0].name)
                  for op in graph.get_operations()
                  if op.type == 'Identity' and op.inputs[0].op.type == 'Const'}

    for node in graph_def.node:
        for i in range(len(node.input)):
            replacement = removables.get(_op_name_from_tensor(node.input[i]))
            if replacement:
                node.input[i] = replacement

    new_graph_def = GraphDef()
    for node in graph_def.node:
        if node.name not in removables:
            new_graph_def.node.append(node)

    return new_graph_def


def _eval_candidates(graph):
    evaluables = set()
    changed = True
    while changed:
        changed = False
        for op in graph.get_operations():
            if op not in evaluables and all(tensor.op in evaluables for tensor in op.inputs) and op.type != 'Placeholder':
                evaluables.add(op)
                changed = True

    candidates = set()
    for op in graph.get_operations():
        if op not in evaluables:
            for tensor in op.inputs:
                if tensor.op in evaluables and tensor.op.type != 'Const':
                    candidates.add(tensor)

    return candidates


def _fold_constant_tensors(graph_def):
    graph = import_graph_def(graph_def)

    evaluated = {}
    with tf.Session(graph=graph) as session:
        for tensor in _eval_candidates(graph):
            evaluated[tensor.name] = _try_eval(tensor, session)

    results = {}
    for op in graph.get_operations():
        results[op.name] = [evaluated.get(tensor.name) for tensor in op.outputs]

    remap = {}
    changed = False
    new_graph_def = GraphDef()
    for node in graph_def.node:
        values = results[node.name]
        all_evaluated = all(value is not None for value in values) and len(values) > 0

        for idx, value in enumerate(values):
            if value is not None:
                arg_name = node.name if idx == 0 else node.name + ':{}'.format(idx)
                const_name = node.name if idx == 0 and all_evaluated else node.name + '//{}'.format(idx)
                remap[arg_name] = const_name
                new_graph_def.node.append(_make_const_node(value, const_name))
                changed = True
        if not all_evaluated:
            new_graph_def.node.append(node)

    for node in new_graph_def.node:
        for i in range(len(node.input)):
            remapped = remap.get(node.input[i])
            if remapped is not None:
                node.input[i] = remapped

    return new_graph_def, changed


def _find_reachables_forward(graph, reachables):
    changed = True
    while changed:
        changed = False
        for op in graph.get_operations():
            if op.name not in reachables and any(tensor.op.name in reachables for tensor in op.inputs):
                reachables.add(op.name)
                changed = True
    return reachables


def _find_reachables_backward(graph, reachables):
    changed = True
    while changed:
        changed = False
        for op in reversed(graph.get_operations()):
            if op.name in reachables:
                for tensor in op.inputs:
                    if tensor.op.name not in reachables:
                        reachables.add(tensor.op.name)
                        changed = True
    return reachables


def _retain_nodes(graph_def, node_names):
    new_graph_def = GraphDef()
    for node in graph_def.node:
        if node.name in node_names:
            new_graph_def.node.append(node)

    for node in new_graph_def.node:
        for idx in reversed(range(len(node.input))):
            name = node.input[idx]
            if name[0] == '^' and name[1:] not in node_names:
                del node.input[idx]

    return new_graph_def


def _retain_reachables_from_placeholders(graph_def):
    graph = import_graph_def(graph_def)

    reachables = {op.name for op in graph.get_operations() if op.type == 'Placeholder'}
    if len(reachables) == 0:
        return graph_def

    reachables = _find_reachables_forward(graph, reachables)
    reachables = _find_reachables_backward(graph, reachables)

    return _retain_nodes(graph_def, reachables)


def _op_name_from_tensor(name):
    if name[0] == '^':
        name = name[1:]
    pos = name.find(':')
    if pos != -1 and name[pos+1:].isdigit():
        name = name[:pos]
    return name


def fold_constant_tensors(graph_def):
    graph_def = _remove_const_control_dependencies(graph_def)
    graph_def = _remove_const_identities(graph_def)

    graph_def, changed = _freeze_shape_tensors(graph_def)
    graph_def, changed = _fold_constant_tensors(graph_def)
    while changed:
        graph_def, changed = _freeze_shape_tensors(graph_def)
        if changed:
            graph_def, changed = _fold_constant_tensors(graph_def)

    graph_def = _retain_reachables_from_placeholders(graph_def)

    return reinfer_shapes(graph_def)


def set_input_shapes(graph_def, input_shapes):
    graph = import_graph_def(graph_def)
    placeholders = {op.name: (op.outputs[0].shape, op.outputs[0].dtype)
                    for op in graph.get_operations() if op.type == 'Placeholder'}

    graph = tf.Graph()
    with graph.as_default():
        input_map = {}
        for name, shape in six.iteritems(input_shapes):
            if name not in placeholders:
                raise IOError("Model has no input named '{}'".format(name))

            orig_shape, dtype = placeholders[name]
            if orig_shape.rank is not None and len(shape) != orig_shape.rank:
                raise IOError("Shape rank for input '{}' does not match that of the model ({} vs {})"
                              .format(name, len(shape), orig_shape.rank))

            input_map[name] = tf.placeholder(shape=shape, dtype=dtype, name=name)

        for name, (shape, dtype) in placeholders.items():
            if name not in input_map:
                input_map[name] = tf.placeholder(shape=shape, dtype=dtype, name=name)

        tf.import_graph_def(graph_def, name='', input_map=input_map)

    used = {tensor.op.name for op in graph.get_operations() for tensor in op.inputs}

    graph_def = GraphDef()
    for op in graph.get_operations():
        if op.type != 'Placeholder' or op.name in used:
            graph_def.node.append(op.node_def)

    return reinfer_shapes(graph_def)


def retain_reachables_from_outputs(graph_def, output_names):
    graph = import_graph_def(graph_def)

    reachables = _find_reachables_backward(graph, set(output_names))

    return _retain_nodes(graph_def, reachables)


def insert_rename_identities(graph_def, tensor_rename):
    for tensor, name in six.iteritems(tensor_rename):
        tensor_name = _remove_zero_index(tensor.name)
        if name != tensor_name:
            graph_def.node.append(_make_identity_node(tensor_name, name,
                                                      tensor.dtype.as_numpy_dtype,
                                                      tuple(tensor.shape.as_list())))
    return graph_def


def check_finite(graph_def):
    for node in graph_def.node:
        attribs = _get_attributes(node.attr)
        for key, value in six.iteritems(attribs):
            if isinstance(value, np.ndarray) and np.issubdtype(value.dtype, np.number) and not np.all(np.isfinite(value)):
                raise ValueError("Attribute '{}' of op '{}' named '{}' contains nan or inf".
                                 format(key, node.op, node.name))


def check_variables(session):
    variables = tf.global_variables()
    for variable in variables:
        value = session.run(variable)
        if np.issubdtype(value.dtype, np.number) and not np.all(np.isfinite(value)):
            raise ValueError("Variable '{}' contains nan or inf".format(variable.name))



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: tflite

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class NonMaxSuppressionV5Options(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsNonMaxSuppressionV5Options(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = NonMaxSuppressionV5Options()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def NonMaxSuppressionV5OptionsBufferHasIdentifier(cls, buf, offset, size_prefixed=False):
        return flatbuffers.util.BufferHasIdentifier(buf, offset, b"\x54\x46\x4C\x33", size_prefixed=size_prefixed)

    # NonMaxSuppressionV5Options
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

def NonMaxSuppressionV5OptionsStart(builder): builder.StartObject(0)
def NonMaxSuppressionV5OptionsEnd(builder): return builder.EndObject()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: tflite

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class BidirectionalSequenceLSTMOptions(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsBidirectionalSequenceLSTMOptions(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = BidirectionalSequenceLSTMOptions()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def BidirectionalSequenceLSTMOptionsBufferHasIdentifier(cls, buf, offset, size_prefixed=False):
        return flatbuffers.util.BufferHasIdentifier(buf, offset, b"\x54\x46\x4C\x33", size_prefixed=size_prefixed)

    # BidirectionalSequenceLSTMOptions
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # BidirectionalSequenceLSTMOptions
    def FusedActivationFunction(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int8Flags, o + self._tab.Pos)
        return 0

    # BidirectionalSequenceLSTMOptions
    def CellClip(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float32Flags, o + self._tab.Pos)
        return 0.0

    # BidirectionalSequenceLSTMOptions
    def ProjClip(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float32Flags, o + self._tab.Pos)
        return 0.0

    # BidirectionalSequenceLSTMOptions
    def MergeOutputs(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False

    # BidirectionalSequenceLSTMOptions
    def TimeMajor(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return True

    # BidirectionalSequenceLSTMOptions
    def AsymmetricQuantizeInputs(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False

def BidirectionalSequenceLSTMOptionsStart(builder): builder.StartObject(6)
def BidirectionalSequenceLSTMOptionsAddFusedActivationFunction(builder, fusedActivationFunction): builder.PrependInt8Slot(0, fusedActivationFunction, 0)
def BidirectionalSequenceLSTMOptionsAddCellClip(builder, cellClip): builder.PrependFloat32Slot(1, cellClip, 0.0)
def BidirectionalSequenceLSTMOptionsAddProjClip(builder, projClip): builder.PrependFloat32Slot(2, projClip, 0.0)
def BidirectionalSequenceLSTMOptionsAddMergeOutputs(builder, mergeOutputs): builder.PrependBoolSlot(3, mergeOutputs, 0)
def BidirectionalSequenceLSTMOptionsAddTimeMajor(builder, timeMajor): builder.PrependBoolSlot(4, timeMajor, 1)
def BidirectionalSequenceLSTMOptionsAddAsymmetricQuantizeInputs(builder, asymmetricQuantizeInputs): builder.PrependBoolSlot(5, asymmetricQuantizeInputs, 0)
def BidirectionalSequenceLSTMOptionsEnd(builder): return builder.EndObject()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: tflite

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class ReshapeOptions(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsReshapeOptions(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = ReshapeOptions()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def ReshapeOptionsBufferHasIdentifier(cls, buf, offset, size_prefixed=False):
        return flatbuffers.util.BufferHasIdentifier(buf, offset, b"\x54\x46\x4C\x33", size_prefixed=size_prefixed)

    # ReshapeOptions
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # ReshapeOptions
    def NewShape(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(flatbuffers.number_types.Int32Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return 0

    # ReshapeOptions
    def NewShapeAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Int32Flags, o)
        return 0

    # ReshapeOptions
    def NewShapeLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # ReshapeOptions
    def NewShapeIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        return o == 0

def ReshapeOptionsStart(builder): builder.StartObject(1)
def ReshapeOptionsAddNewShape(builder, newShape): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(newShape), 0)
def ReshapeOptionsStartNewShapeVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def ReshapeOptionsEnd(builder): return builder.EndObject()



# Copyright (c) 2020 The Khronos Group Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import numpy as np
import nnef
import sys
import os


def _is_lambda(value):
    LAMBDA = lambda: 0
    return isinstance(value, type(LAMBDA)) and value.__name__ == LAMBDA.__name__


def _ensure_lambda(value):
    return value() if not _is_lambda(value) else value


def uniform(min=0.0, max=1.0):
    return lambda shape: np.random.uniform(min, max, shape).astype(np.float32)


def normal(mean=0.0, std=1.0):
    return lambda shape: np.random.normal(mean, std, shape).astype(np.float32)


def bernoulli(prob=0.5):
    return lambda shape: np.random.uniform(0.0, 1.0, shape) > prob


def integers(min=0, max=100):
    return lambda shape: np.random.randint(min, max, shape).astype(np.int32)


def main(args):
    if args.seed is not None:
        np.random.seed(args.seed)

    distributions = {
        'scalar': uniform(0.0, 1.0),
        'integer': integers(0, 100),
        'logical': bernoulli(0.5),
    }

    try:
        random = eval(args.random)
        if isinstance(random, dict):
            distributions.update({key: _ensure_lambda(value) for key, value in random.items()})
        else:
            random = _ensure_lambda(random)
            if args.random.startswith('integers'):
                distributions['integer'] = random
            elif args.random.startswith('bernoulli'):
                distributions['logical'] = random
            else:
                distributions['scalar'] = random
    except Exception as e:
        print("Could not evaluate distribution: " + str(e), file=sys.stderr)
        return -1

    graph = nnef.parse_file(os.path.join(args.model, 'graph.nnef'))

    for op in graph.operations:
        if args.weights and op.name == 'variable':
            label = op.attribs['label']
            shape = op.attribs['shape']
            data = distributions[op.dtype](shape)
            filename = os.path.join(args.model, label + '.dat')

            os.makedirs(os.path.split(filename)[0], exist_ok=True)
            with open(filename, 'wb') as file:
                nnef.write_tensor(file, data)

            if args.verbose:
                print("Generated weight '{}'".format(filename))

        if args.inputs and op.name == 'external':
            name = op.outputs['output']
            shape = op.attribs['shape']
            data = distributions[op.dtype](shape)
            filename = os.path.join(args.model, args.inputs, name + '.dat')

            os.makedirs(os.path.split(filename)[0], exist_ok=True)
            with open(filename, 'wb') as file:
                nnef.write_tensor(file, data)

            if args.verbose:
                print("Generated input '{}'".format(filename))

    return 0


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('model', type=str,
                        help='The model to generate')
    parser.add_argument('--random', type=str, required=True,
                        help='Random distribution for input generation, possibly per dtype')
    parser.add_argument('--seed', type=int, default=None,
                        help='Random seed for input generation')
    parser.add_argument('--weights', action='store_true',
                        help='Generate weights')
    parser.add_argument('--inputs', type=str, nargs='?', default=None, const='.',
                        help='Generate inputs')
    parser.add_argument('--verbose', action='store_true',
                        help='Weather to print generated file names')
    exit(main(parser.parse_args()))



# Copyright (c) 2020 The Khronos Group Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .protobuf import GraphDef, NodeDef
from .writer import _build_attribute
from .utils import import_graph_def
try:
    import tensorflow.compat.v1 as tf
except ImportError:
    import tensorflow as tf
from collections.abc import Sequence
import inspect


class _Composite:

    instances = []

    def __init__(self, id, func, attribs, inputs, outputs):
        self.id = id
        self.func = func
        self.attribs = attribs
        self.inputs = inputs
        self.outputs = outputs

    @staticmethod
    def function(func):
        def wrapper(*args, **kwargs):
            results = func(*args, **kwargs)

            name = kwargs.get('name')
            if name is not None:
                del kwargs['name']
            id = name or len(_Composite.instances)

            signature = inspect.signature(func)
            bound = signature.bind(*args, **kwargs)
            bound.apply_defaults()

            attribs = {name: value for name, value in bound.arguments.items()
                       if not isinstance(value, tf.Tensor) and value is not None}
            inputs = [value for value in bound.arguments.values()
                      if isinstance(value, tf.Tensor)]
            outputs = (results,) if not isinstance(results, (list, tuple)) else results

            assert all(isinstance(value, tf.Tensor) for value in outputs), \
                "Results of composite function must be tensors"
            assert not any(tensor in inputs for tensor in outputs), \
                "Results of composite function cannot be input arguments at the same time"

            _Composite.instances.append(_Composite(id, func, attribs, inputs, outputs))

            return results
        return wrapper

    @property
    def name(self):
        return self.id if isinstance(self.id, str) else 'Composite' + str(self.id)


function = _Composite.function


def _is_tensor(value):
    return isinstance(value, tf.Tensor) or (isinstance(value, Sequence) and all(_is_tensor(item) for item in value))


def _node_name_from_tensor(name):
    if name[0] == '^':
        name = name[1:]
    pos = name.find(':')
    if pos != -1 and name[pos+1:].isdigit():
        name = name[:pos]
    return name


def _input_name_from_tensor(name):
    return name[:-2] if name.endswith(':0') else name


def _build_node_def(composite):
    node_def = NodeDef()
    node_def.op = 'PyFunc'
    node_def.name = composite.name
    node_def.input.extend([_input_name_from_tensor(arg.name) for arg in composite.inputs])

    input_dtypes = [tensor.dtype.as_numpy_dtype for tensor in composite.inputs]
    output_dtypes = [tensor.dtype.as_numpy_dtype for tensor in composite.outputs]
    output_shapes = [tuple(tensor.shape.as_list()) for tensor in composite.outputs]

    _build_attribute(node_def.attr['Tin'], input_dtypes)
    _build_attribute(node_def.attr['Tout'], output_dtypes)
    _build_attribute(node_def.attr['token'], composite.func.__name__)
    _build_attribute(node_def.attr['_output_shapes'], output_shapes)
    for name, value in composite.attribs.items():
        _build_attribute(node_def.attr['_$' + name + '$_'], value)
    return node_def


def _remap_tensors(tensors, graph):
    return type(tensors)(graph.get_tensor_by_name(tensor.name) for tensor in tensors)


def _tensor_producers_and_consumers(graph):
    producers_and_consumers = {tensor: [tensor.op] for op in graph.get_operations() for tensor in op.outputs}
    for op in graph.get_operations():
        for tensor in op.inputs:
            ops = producers_and_consumers[tensor]
            if op not in ops:
                ops.append(op)

    return producers_and_consumers


def _find_subgraph(composite, producers_and_consumers):
    queue = [tensor.op for tensor in composite.outputs]
    subgraph = {item.name for item in queue}

    idx = 0
    while idx < len(queue):
        op = queue[idx]
        idx += 1
        tensors = [tensor for tensor in op.inputs if tensor not in composite.inputs] + \
                  [tensor for tensor in op.outputs if tensor not in composite.outputs]
        for tensor in tensors:
            for op in producers_and_consumers[tensor]:
                if op.name not in subgraph:
                    subgraph.add(op.name)
                    queue.append(op)
    return subgraph


def replace_composites_with_py_functions(graph_def):
    graph = import_graph_def(graph_def)
    for composite in _Composite.instances:
        composite.inputs = _remap_tensors(composite.inputs, graph)
        composite.outputs = _remap_tensors(composite.outputs, graph)

    producers_and_consumers = _tensor_producers_and_consumers(graph)

    tensor_remap = {}
    subgraph_ops = set()
    for composite in _Composite.instances:
        subgraph_ops.update(_find_subgraph(composite, producers_and_consumers))

        for idx, tensor in enumerate(composite.outputs):
            tensor_remap[_input_name_from_tensor(tensor.name)] = \
                composite.name + ':' + str(idx) if idx > 0 else composite.name

    new_graph_def = GraphDef()
    for node in graph_def.node:
        if node.name not in subgraph_ops:
            new_graph_def.node.append(node)

    for composite in _Composite.instances:
        new_graph_def.node.append(_build_node_def(composite))

    for node in new_graph_def.node:
        for i in range(len(node.input)):
            remapped = tensor_remap.get(node.input[i])
            if remapped is not None:
                node.input[i] = remapped

    return new_graph_def


def reset_composites():
    _Composite.instances = []



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: tflite

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class QuantizationParameters(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsQuantizationParameters(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = QuantizationParameters()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def QuantizationParametersBufferHasIdentifier(cls, buf, offset, size_prefixed=False):
        return flatbuffers.util.BufferHasIdentifier(buf, offset, b"\x54\x46\x4C\x33", size_prefixed=size_prefixed)

    # QuantizationParameters
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # QuantizationParameters
    def Min(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(flatbuffers.number_types.Float32Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return 0

    # QuantizationParameters
    def MinAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Float32Flags, o)
        return 0

    # QuantizationParameters
    def MinLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # QuantizationParameters
    def MinIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        return o == 0

    # QuantizationParameters
    def Max(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(flatbuffers.number_types.Float32Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return 0

    # QuantizationParameters
    def MaxAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Float32Flags, o)
        return 0

    # QuantizationParameters
    def MaxLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # QuantizationParameters
    def MaxIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        return o == 0

    # QuantizationParameters
    def Scale(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(flatbuffers.number_types.Float32Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return 0

    # QuantizationParameters
    def ScaleAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Float32Flags, o)
        return 0

    # QuantizationParameters
    def ScaleLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # QuantizationParameters
    def ScaleIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        return o == 0

    # QuantizationParameters
    def ZeroPoint(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(flatbuffers.number_types.Int64Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 8))
        return 0

    # QuantizationParameters
    def ZeroPointAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Int64Flags, o)
        return 0

    # QuantizationParameters
    def ZeroPointLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # QuantizationParameters
    def ZeroPointIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        return o == 0

    # QuantizationParameters
    def DetailsType(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 0

    # QuantizationParameters
    def Details(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            from flatbuffers.table import Table
            obj = Table(bytearray(), 0)
            self._tab.Union(obj, o)
            return obj
        return None

    # QuantizationParameters
    def QuantizedDimension(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

def QuantizationParametersStart(builder): builder.StartObject(7)
def QuantizationParametersAddMin(builder, min): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(min), 0)
def QuantizationParametersStartMinVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def QuantizationParametersAddMax(builder, max): builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(max), 0)
def QuantizationParametersStartMaxVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def QuantizationParametersAddScale(builder, scale): builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(scale), 0)
def QuantizationParametersStartScaleVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def QuantizationParametersAddZeroPoint(builder, zeroPoint): builder.PrependUOffsetTRelativeSlot(3, flatbuffers.number_types.UOffsetTFlags.py_type(zeroPoint), 0)
def QuantizationParametersStartZeroPointVector(builder, numElems): return builder.StartVector(8, numElems, 8)
def QuantizationParametersAddDetailsType(builder, detailsType): builder.PrependUint8Slot(4, detailsType, 0)
def QuantizationParametersAddDetails(builder, details): builder.PrependUOffsetTRelativeSlot(5, flatbuffers.number_types.UOffsetTFlags.py_type(details), 0)
def QuantizationParametersAddQuantizedDimension(builder, quantizedDimension): builder.PrependInt32Slot(6, quantizedDimension, 0)
def QuantizationParametersEnd(builder): return builder.EndObject()



# Copyright (c) 2017 The Khronos Group Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .AbsOptions import *
from .ActivationFunctionType import *
from .AddNOptions import *
from .AddOptions import *
from .ArgMaxOptions import *
from .ArgMinOptions import *
from .BatchToSpaceNDOptions import *
from .BidirectionalSequenceLSTMOptions import *
from .BidirectionalSequenceRNNOptions import *
from .Buffer import *
from .BuiltinOperator import *
from .BuiltinOptions import *
from .CallOptions import *
from .CastOptions import *
from .CombinerType import *
from .ConcatEmbeddingsOptions import *
from .ConcatenationOptions import *
from .Conv2DOptions import *
from .CosOptions import *
from .CustomOptionsFormat import *
from .CustomQuantization import *
from .DepthToSpaceOptions import *
from .DepthwiseConv2DOptions import *
from .DequantizeOptions import *
from .DivOptions import *
from .EmbeddingLookupSparseOptions import *
from .EqualOptions import *
from .ExpandDimsOptions import *
from .ExpOptions import *
from .FakeQuantOptions import *
from .FillOptions import *
from .FloorDivOptions import *
from .FloorModOptions import *
from .FullyConnectedOptions import *
from .FullyConnectedOptionsWeightsFormat import *
from .GatherNdOptions import *
from .GatherOptions import *
from .GreaterEqualOptions import *
from .GreaterOptions import *
from .HardSwishOptions import *
from .IfOptions import *
from .L2NormOptions import *
from .LeakyReluOptions import *
from .LessEqualOptions import *
from .LessOptions import *
from .LocalResponseNormalizationOptions import *
from .LogicalAndOptions import *
from .LogicalNotOptions import *
from .LogicalOrOptions import *
from .LogSoftmaxOptions import *
from .LSHProjectionOptions import *
from .LSHProjectionType import *
from .LSTMKernelType import *
from .LSTMOptions import *
from .MatrixDiagOptions import *
from .MatrixSetDiagOptions import *
from .MaximumMinimumOptions import *
from .Metadata import *
from .MirrorPadMode import *
from .MirrorPadOptions import *
from .Model import *
from .MulOptions import *
from .NegOptions import *
from .NonMaxSuppressionV4Options import *
from .NonMaxSuppressionV5Options import *
from .NotEqualOptions import *
from .OneHotOptions import *
from .OperatorCode import *
from .Operator import *
from .PackOptions import *
from .Padding import *
from .PadOptions import *
from .PadV2Options import *
from .Pool2DOptions import *
from .PowOptions import *
from .QuantizationDetails import *
from .QuantizationParameters import *
from .QuantizeOptions import *
from .RangeOptions import *
from .RankOptions import *
from .ReducerOptions import *
from .ReshapeOptions import *
from .ResizeBilinearOptions import *
from .ResizeNearestNeighborOptions import *
from .ReverseSequenceOptions import *
from .ReverseV2Options import *
from .RNNOptions import *
from .ScatterNdOptions import *
from .SelectOptions import *
from .SequenceRNNOptions import *
from .ShapeOptions import *
from .SkipGramOptions import *
from .SliceOptions import *
from .SoftmaxOptions import *
from .SpaceToBatchNDOptions import *
from .SpaceToDepthOptions import *
from .SparseToDenseOptions import *
from .SplitOptions import *
from .SplitVOptions import *
from .SquaredDifferenceOptions import *
from .SquareOptions import *
from .SqueezeOptions import *
from .StridedSliceOptions import *
from .SubGraph import *
from .SubOptions import *
from .SVDFOptions import *
from .Tensor import *
from .TensorType import *
from .TileOptions import *
from .TopKV2Options import *
from .TransposeConvOptions import *
from .TransposeOptions import *
from .UnidirectionalSequenceLSTMOptions import *
from .UniqueOptions import *
from .UnpackOptions import *
from .WhereOptions import *
from .WhileOptions import *
from .ZerosLikeOptions import *
from .BatchMatMulOptions import *
from .DensifyOptions import *
from .SegmentSumOptions import *
from .SelectV2Options import *



# Copyright (c) 2020 The Khronos Group Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import division, print_function, absolute_import

from .helpers import *
from ....model import *
try:
    from flatbuffers import flexbuffers
    has_flexbuffers = True
except ImportError:
    has_flexbuffers = False
import sys
import six


def _get_quantization(tensor):
    quant = tensor.Quantization()
    if quant is None:
        return None

    if quant.MinLength() == 0:
        min = None
    elif quant.MinLength() == 1:
        min = float(quant.Min(0))
    else:
        min = quant.MinAsNumpy()

    if quant.MaxLength() == 0:
        max = None
    elif quant.MaxLength() == 1:
        max = float(quant.Max(0))
    else:
        max = quant.MaxAsNumpy()

    if quant.ScaleLength() == 0:
        scale = None
    elif quant.ScaleLength() == 1:
        scale = float(quant.Scale(0))
    else:
        scale = quant.ScaleAsNumpy()

    if quant.ZeroPointLength() == 0:
        zero_point = None
    elif quant.ZeroPointLength() == 1:
        zero_point = int(quant.ZeroPoint(0))
    else:
        zero_point = quant.ZeroPointAsNumpy()

    if all(x is None for x in [min, max, scale, zero_point]):
        return None
    else:
        return dict(min=min, max=max, zero_point=zero_point, scale=scale)


def _get_data_as_ndarray(buffer, dtype, shape):
    return buffer.DataAsNumpy().view(dtype).reshape(shape) if buffer.DataLength() != 0 else None


def _get_options_starter_ender(optionsClass):
    className = optionsClass.__name__
    optionsModule = sys.modules[optionsClass.__module__]
    moduleDict = optionsModule.__dict__
    return moduleDict[className + 'Start'], moduleDict[className + 'End']


def _enumerate_attributes(optionsClass, optionsObject):
    getters = enumerate_options_getters(optionsClass)
    length_getters = enumerate_options_length_getters(optionsClass)

    attribs = {}
    for name, getter in getters.items():
        length_getter = length_getters.get(name)

        value = getter(optionsObject) if length_getter is None else \
            [getter(optionsObject, i) for i in range(length_getter(optionsObject))]

        attribs[name] = substitute_enum_value_with_name(name, value, optionsClass)

    return attribs


def _decode_custom_options(bytes):
    root = flexbuffers.GetRoot(bytes)
    assert root.IsMap
    return {key: value for key, value in six.iteritems(root.AsMap.Value) if not key.startswith('_')}


def read_flatbuffers(filename):
    with open(filename, 'rb') as file:
        bytes = bytearray(file.read())

    model = fb.Model.GetRootAsModel(bytes, 0)

    if model.SubgraphsLength() != 1:
        raise NotImplementedError('graphs with multiple sub-graphs are not supported')

    subgraph = model.Subgraphs(0)
    name = subgraph.Name()

    graph = Graph(name.decode() if name is not None else None)

    tensors = []
    for i in range(subgraph.TensorsLength()):
        tensor = subgraph.Tensors(i)
        name = tensor.Name().decode()
        shape = tuple(tensor.Shape(i) for i in range(tensor.ShapeLength()))
        dtype = DtypeToNumpy[tensor.Type()]
        buffer = model.Buffers(tensor.Buffer())
        data = _get_data_as_ndarray(buffer, dtype, shape)
        quant = _get_quantization(tensor)
        tensors.append(Tensor(graph, name, shape, dtype, data, quant))

    for i in range(subgraph.OperatorsLength()):
        operator = subgraph.Operators(i)
        operatorCode = model.OperatorCodes(operator.OpcodeIndex())
        builtinCode = operatorCode.BuiltinCode()
        opType = BuiltinOperatorTypeByValue[builtinCode] if builtinCode != fb.BuiltinOperator.CUSTOM else \
            operatorCode.CustomCode().decode('ascii')
        custom = builtinCode == fb.BuiltinOperator.CUSTOM

        options = operator.BuiltinOptions()
        optionsClass = BuiltinOptionsClasses[operator.BuiltinOptionsType()]

        inputs = [tensors[operator.Inputs(i)] for i in range(operator.InputsLength()) if operator.Inputs(i) != -1]
        outputs = [tensors[operator.Outputs(i)] for i in range(operator.OutputsLength()) if operator.Outputs(i) != -1]

        if options is not None and optionsClass is not None:
            optionsObject = optionsClass()
            optionsObject.Init(options.Bytes, options.Pos)
            attribs = _enumerate_attributes(optionsClass, optionsObject)
        elif custom:
            bytes = operator.CustomOptionsAsNumpy().tobytes()
            attribs = _decode_custom_options(bytes) if has_flexbuffers else {CustomOptionsKey: bytes}
        else:
            attribs = {}

        Operation(graph, type=opType, custom=custom, attribs=attribs, inputs=tuple(inputs), outputs=tuple(outputs))

    inputs = []
    for i in range(subgraph.InputsLength()):
        tensor_index = subgraph.Inputs(i)
        inputs.append(tensors[tensor_index])

    outputs = []
    for i in range(subgraph.OutputsLength()):
        tensor_index = subgraph.Outputs(i)
        outputs.append(tensors[tensor_index])

    graph.inputs = inputs
    graph.outputs = outputs

    return graph


class Reader(object):

    def __call__(self, filename):
        return read_flatbuffers(filename)



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: tflite

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class UnidirectionalSequenceLSTMOptions(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsUnidirectionalSequenceLSTMOptions(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = UnidirectionalSequenceLSTMOptions()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def UnidirectionalSequenceLSTMOptionsBufferHasIdentifier(cls, buf, offset, size_prefixed=False):
        return flatbuffers.util.BufferHasIdentifier(buf, offset, b"\x54\x46\x4C\x33", size_prefixed=size_prefixed)

    # UnidirectionalSequenceLSTMOptions
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # UnidirectionalSequenceLSTMOptions
    def FusedActivationFunction(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int8Flags, o + self._tab.Pos)
        return 0

    # UnidirectionalSequenceLSTMOptions
    def CellClip(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float32Flags, o + self._tab.Pos)
        return 0.0

    # UnidirectionalSequenceLSTMOptions
    def ProjClip(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float32Flags, o + self._tab.Pos)
        return 0.0

    # UnidirectionalSequenceLSTMOptions
    def TimeMajor(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False

    # UnidirectionalSequenceLSTMOptions
    def AsymmetricQuantizeInputs(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False

def UnidirectionalSequenceLSTMOptionsStart(builder): builder.StartObject(5)
def UnidirectionalSequenceLSTMOptionsAddFusedActivationFunction(builder, fusedActivationFunction): builder.PrependInt8Slot(0, fusedActivationFunction, 0)
def UnidirectionalSequenceLSTMOptionsAddCellClip(builder, cellClip): builder.PrependFloat32Slot(1, cellClip, 0.0)
def UnidirectionalSequenceLSTMOptionsAddProjClip(builder, projClip): builder.PrependFloat32Slot(2, projClip, 0.0)
def UnidirectionalSequenceLSTMOptionsAddTimeMajor(builder, timeMajor): builder.PrependBoolSlot(3, timeMajor, 0)
def UnidirectionalSequenceLSTMOptionsAddAsymmetricQuantizeInputs(builder, asymmetricQuantizeInputs): builder.PrependBoolSlot(4, asymmetricQuantizeInputs, 0)
def UnidirectionalSequenceLSTMOptionsEnd(builder): return builder.EndObject()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: tflite

class ActivationFunctionType(object):
    NONE = 0
    RELU = 1
    RELU_N1_TO_1 = 2
    RELU6 = 3
    TANH = 4
    SIGN_BIT = 5




# Copyright (c) 2020 The Khronos Group Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
from functools import reduce
from .io.nnef import Reader


def _volume(shape):
    return reduce(lambda x, y: x * y, shape, 1)


def _count_macs(op, include_pooling, include_upsampling, include_normalization, include_reduction):
    if len(op.inputs) == 0 or len(op.outputs) == 0:
        return 0
    
    input_volume = _volume(op.inputs[0].shape)
    output_volume = _volume(op.outputs[0].shape)

    if op.type in ['conv', 'deconv']:
        volume = input_volume if op.type == 'deconv' else output_volume
        filter_shape = op.inputs[1].shape
        return volume * _volume(filter_shape[1:])
    elif op.type in ['separable_conv', 'separable_deconv']:
        volume = input_volume if op.type == 'separable_deconv' else output_volume
        filter_shape = op.inputs[1].shape
        inter_channels = filter_shape[0]
        inter_volume = output_volume / op.outputs[0].shape[1] * inter_channels if op.type == 'separable_deconv' else \
                        input_volume / op.inputs[0].shape[1] * inter_channels
        return inter_volume * _volume(filter_shape[2:]) + volume * inter_channels
    elif op.type == 'linear':
        filter_shape = op.inputs[1].shape
        return output_volume * filter_shape[-1]
    elif op.type == 'matmul':
        filter_shape = op.inputs[1].shape
        return output_volume * (filter_shape[-1] if op.attribs['transposeB'] else filter_shape[-2])
    elif op.type in ['max_pool', 'avg_pool', 'rms_pool', 'max_pool_with_index', 'box', 'debox'] and include_pooling:
        volume = input_volume if op.type == 'debox' else output_volume
        kernel_size = op.attribs['size']
        return volume * _volume(kernel_size)
    elif op.type == 'multilinear_upsample' and include_upsampling:
        factor = op.attribs['factor']
        method = op.attribs['method']
        if method == 'symmetric':
            kernel_size = [2 * f for f in factor]
        elif method == 'asymmetric':
            kernel_size = [2 * f - 1 for f in factor]
        else:
            kernel_size = factor
        return input_volume * _volume(kernel_size)
    elif op.type in ['local_response_normalization', 'local_mean_normalization',
                     'local_variance_normalization', 'local_contrast_normalization'] and include_normalization:
        kernel_size = op.attribs['size']
        return output_volume * _volume(kernel_size)
    elif op.type in ['l1_normalization', 'l2_normalization', 'batch_normalization'] and include_normalization:
        return output_volume
    elif op.type in ['sum_reduce', 'max_reduce', 'min_reduce',
                     'mean_reduce', 'all_reduce', 'any_reduce'] and include_reduction:
        return input_volume
    else:
        return 0


def get_custom_shapes(module_names):
    import importlib

    CUSTOM_SHAPES = "CUSTOM_SHAPES"

    shapes = {}
    for module_name in module_names:
        module = importlib.import_module(module_name)
        if hasattr(module, CUSTOM_SHAPES):
            shapes.update(getattr(module, CUSTOM_SHAPES))

    return shapes


def main(args):
    custom_shapes = get_custom_shapes(args.custom_shapes) if args.custom_shapes is not None else None
    reader = Reader(infer_shapes=True, custom_shapes=custom_shapes)
    graph = reader(args.model)

    macs = 0
    for op in graph.operations:
        macs += _count_macs(op, args.include_pooling, args.include_upsampling,
                            args.include_normalization, args.include_reduction)

    volume = 0
    for tensor in graph.tensors:
        volume += _volume(tensor.shape)

    gmacs = macs / 1000 / 1000 / 1000
    mbytes = volume * 4 / 1000 / 1000
    print('GMACs = {}'.format(gmacs))
    print('Total memory in Mbytes (supposing float32) = {}'.format(mbytes))
    return 0


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('model', type=str,
                        help='The model to visualize')
    parser.add_argument('--include-pooling', action='store_true',
                        help='Whether to include pooling operations in the calculation')
    parser.add_argument('--include-upsampling', action='store_true',
                        help='Whether to include (linear) upsampling operations in the calculation')
    parser.add_argument('--include-normalization', action='store_true',
                        help='Whether to include normalization operations in the calculation')
    parser.add_argument('--include-reduction', action='store_true',
                        help='Whether to include reduction operations in the calculation')
    parser.add_argument('--custom-shapes', type=str, nargs='+',
                        help='Module(s) containing custom shape inference code (when converting to NNEF)')
    exit(main(parser.parse_args()))



# Copyright (c) 2020 The Khronos Group Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import division, print_function, absolute_import

from collections.abc import Sequence
from functools import reduce

import six
import typing
import numpy as np


# noinspection PyProtectedMember
class Tensor:

    def __init__(self,
                 graph,             # type: Graph,
                 name=None,         # type: typing.Optional[str]
                 shape=None,        # type: typing.Optional[typing.Tuple[int, ...]]
                 dtype=None,        # type: typing.Optional[type]
                 data=None,         # type: typing.Union[None, np.ndarray, typing.Any]
                 quant=None         # type: typing.Optional[typing.Dict[str, typing.Any]]
                 ):
        # type: (...)->None
        self._graph = graph
        self._producers = []
        self._consumers = []

        self.name = name            # type: typing.Optional[str]
        self.shape = shape          # type: typing.Optional[typing.Tuple[int, ...]]
        self.dtype = dtype          # type: typing.Optional[type]
        self.data = data            # type: typing.Union[None, np.ndarray, typing.Any]
        self.quant = quant or {}    # type: typing.Optional[typing.Dict[str, typing.Any]]

        assert isinstance(graph, Graph)
        graph._tensors.append(self)

    def copy_with(self, graph=None, name=None, dtype=None, shape=None, data=None, quant=None):
        return Tensor(graph=graph if graph is not None else self.graph,
                      name=name if name is not None else self.name,
                      dtype=dtype if dtype is not None else self.dtype,
                      shape=shape if shape is not None else self.shape,
                      data=data if data is not None else self.data,
                      quant=quant if quant is not None else self.quant)

    @property
    def graph(self):
        # type: ()->typing.Optional[Graph]
        return self._graph

    @property
    def has_producer(self):
        return len(self._producers) != 0

    @property
    def producers(self):
        # type: ()->typing.List[Operation]
        return self._producers

    @property
    def producer(self):
        # type: ()->typing.Optional[Operation]
        assert len(self._producers) <= 1
        return self._producers[0] if len(self._producers) == 1 else None

    @property
    def has_consumer(self):
        return len(self._consumers) != 0

    @property
    def consumers(self):
        # type: ()->typing.List[Operation]
        return self._consumers

    @property
    def consumer(self):
        # type: ()->typing.Optional[Operation]
        return self._consumers[0] if len(self._consumers) == 1 else None

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name):
        assert name is None or isinstance(name, str)
        self._name = name

    @property
    def shape(self):
        return self._shape

    @shape.setter
    def shape(self, shape):
        assert shape is None or isinstance(shape, (list, tuple))
        assert shape is None or all(s is None or isinstance(s, int) for s in shape)
        self._shape = tuple(shape) if shape is not None else None

    @property
    def dtype(self):
        return self._dtype

    @dtype.setter
    def dtype(self, dtype):
        assert dtype is None or isinstance(dtype, type)
        self._dtype = dtype

    @property
    def rank(self):
        # type: ()->typing.Optional[int]
        return len(self.shape) if self.shape is not None else None

    @property
    def volume(self):
        # type: ()->typing.Optional[int]
        return reduce((lambda x, y: x * y), self.shape) if self.shape is not None and \
                                                           all(s is not None for s in self.shape) else None

    @property
    def is_constant(self):
        # type: ()->bool
        return self.data is not None

    def __repr__(self):
        return self.name if self.name is not None else _hex_id(self)

    def __str__(self):
        return '{name}: {dtype}[{shape}]'.format(
            name=self.name if self.name is not None else _hex_id(self),
            dtype=self.dtype.__name__,
            shape=', '.join(str(s) for s in self.shape) if self.shape else '...')


_TensorListOrTupleT = typing.Union[typing.List[Tensor], typing.Tuple[Tensor, ...]]


# noinspection PyProtectedMember
class Operation:

    def __init__(self,
                 graph,             # type: Graph
                 type=None,         # type: typing.Optional[str]
                 name=None,         # type: typing.Optional[str]
                 attribs=None,      # type: typing.Dict[str, typing.Any]
                 inputs=None,       # type: typing.Union[None, Tensor, _TensorListOrTuple]
                 outputs=None,      # type: typing.Union[None, Tensor, _TensorListOrTuple]
                 custom=False,      # type: bool
                 ):
        # type:(...)->None
        self._graph = graph
        self._inputs = tuple()
        self._outputs = tuple()

        assert name is None or isinstance(name, str)
        if attribs is not None:
            assert isinstance(attribs, dict)
            assert all(isinstance(key, str) for key in six.iterkeys(attribs))
            assert all(not isinstance(value, Tensor) for value in six.itervalues(attribs))

        self.type = type                # type: typing.Optional[str]
        self.name = name                # type: typing.Optional[str]
        self.attribs = attribs or {}    # type: typing.Dict[str, typing.Any]
        self.custom = custom            # type: bool

        assert isinstance(graph, Graph)
        graph._operations.append(self)

        if inputs is not None:
            self.inputs = inputs
        if outputs is not None:
            self.outputs = outputs

    def copy_with(self, graph=None, type=None, name=None, attribs=None, inputs=None, outputs=None, custom=None):
        return Operation(graph=graph if graph is not None else self.graph,
                         type=type if type is not None else self.type,
                         name=name if name is not None else self.name,
                         attribs=attribs if attribs is not None else self.attribs,
                         inputs=inputs if inputs is not None else self.inputs,
                         outputs=outputs if outputs is not None else self.outputs,
                         custom=custom if custom is not None else self.custom)

    @property
    def graph(self):
        # type: ()->typing.Optional[Graph]
        return self._graph

    @property
    def inputs(self):
        # type: ()->_TensorListOrTupleT
        return self._inputs

    @property
    def input(self):
        # type: ()->Tensor
        assert len(self._inputs) == 1
        return self._inputs[0]

    @inputs.setter
    def inputs(self, tensors):
        # type: (typing.Union[Tensor, _TensorListOrTupleT])->None
        if isinstance(tensors, Tensor):
            tensors = (tensors,)

        for tensor in self._inputs:
            assert self in tensor._consumers
        for tensor in self._inputs:
            if self in tensor._consumers:
                tensor._consumers.remove(self)

        self._inputs = _ListView(tensors) if isinstance(tensors, list) else tensors
        for tensor in tensors:
            assert isinstance(tensor, Tensor), "got {}".format(type(tensor))
            if self not in tensor._consumers:
                tensor._consumers.append(self)

    @property
    def outputs(self):
        # type: ()->_TensorListOrTupleT
        return self._outputs

    @property
    def output(self):
        # type: ()->Tensor
        assert len(self._outputs) == 1
        return self._outputs[0]

    @outputs.setter
    def outputs(self, tensors):
        # type: (typing.Union[Tensor, _TensorListOrTupleT])->None

        if isinstance(tensors, Tensor):
            tensors = (tensors,)

        for tensor in self._outputs:
            assert self in tensor._producers
            tensor._producers.remove(self)

        self._outputs = _ListView(tensors) if isinstance(tensors, list) else tensors
        for tensor in tensors:
            assert isinstance(tensor, Tensor), "got {}".format(type(tensor))
            assert self not in tensor._producers
            tensor._producers.append(self)

    @property
    def type(self):
        return self._type

    @type.setter
    def type(self, type):
        assert type is None or isinstance(type, str), "got '{}'".format(type)
        self._type = type

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name):
        assert name is None or isinstance(name, str), "got '{}'".format(name)
        self._name = name

    def __repr__(self):
        return self.type if self.type is not None else _hex_id(self)

    def __str__(self):
        return '{outputs} = {op}{{{attribs}}}({inputs})'.format(
            op=self.type if self.type is not None else _hex_id(self),
            inputs=', '.join(repr(tensor) for tensor in self._inputs),
            outputs=', '.join(str(tensor) for tensor in self._outputs),
            attribs=', '.join('{}={}'.format(key, value) for key, value in self.attribs.items()))


# noinspection PyProtectedMember
class Graph:

    def __init__(self, name=None):
        # type:(typing.Optional[str])->None
        self._operations = []
        self._tensors = []
        self._inputs = []
        self._outputs = []
        self._name = name

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name):
        assert name is None or isinstance(name, str)
        self._name = name

    @property
    def operations(self):
        # type: ()->typing.Sequence[Operation]
        return _ListView(self._operations)

    @property
    def tensors(self):
        # type: ()->typing.Sequence[Tensor]
        return _ListView(self._tensors)

    @property
    def inputs(self):
        # type: ()->typing.Sequence[Tensor]
        return _ListView(self._inputs)

    @inputs.setter
    def inputs(self, tensors):
        # type: (_TensorListOrTupleT)->None
        assert isinstance(tensors, (list, tuple))

        self._inputs = tensors

        for tensor in self._inputs:
            assert isinstance(tensor, Tensor)

    @property
    def outputs(self):
        # type: ()->typing.Sequence[Tensor]
        return _ListView(self._outputs)

    @outputs.setter
    def outputs(self, tensors):
        # type: (_TensorListOrTupleT)->None
        assert isinstance(tensors, (list, tuple))

        self._outputs = tensors

        for tensor in self._outputs:
            assert isinstance(tensor, Tensor)

    def remove_tensor(self, tensor):
        # type: (Tensor)->None
        assert len(tensor.producers) == 0
        assert len(tensor.consumers) == 0
        assert tensor not in self._inputs
        assert tensor not in self._outputs
        self._tensors.remove(tensor)
        tensor._graph = None

    def remove_tensors(self, tensors):
        # type: (typing.Iterable[Tensor])->None
        for tensor in tensors:
            assert len(tensor.producers) == 0
            assert len(tensor.consumers) == 0
            assert tensor not in self._inputs
            assert tensor not in self._outputs
        self._tensors = [tensor for tensor in self._tensors if tensor not in tensors]
        for tensor in tensors:
            tensor._graph = None

    def remove_operation(self, operation, unlink=False):
        # type: (Operation, bool)->None
        if unlink:
            operation.inputs = []
            operation.outputs = []
        else:
            assert len(operation.inputs) == 0
            assert len(operation.outputs) == 0
        self._operations.remove(operation)
        operation._graph = None

    def remove_operations(self, operations, unlink=False):
        # type: (typing.Iterable[Operation], bool)->None
        operations = operations if isinstance(operations, set) else set(operations)
        for operation in operations:
            if unlink:
                operation.inputs = []
                operation.outputs = []
            else:
                assert len(operation.inputs) == 0
                assert len(operation.outputs) == 0
        self._operations = [op for op in self._operations if op not in operations]
        for operation in operations:
            operation._graph = None

    def is_unique(self):
        return all(len(t.producers) <= 1 for t in self.tensors)

    def is_sorted(self):
        seen = set()
        for op in self.operations:
            for tensor in op.inputs:
                for producer in tensor.producers:
                    if producer not in seen:
                        return False
            seen.add(op)
        return True

    def sort(self, offset=0):
        count = len(self._operations)
        sorted = {op: False for op in self._operations[offset:]}
        for idx in range(offset, count):
            i = idx
            while i < count and not all(sorted.get(tensor.producer, True) for tensor in self._operations[i].inputs):
                i += 1
            if i == count:  # the graph contains a loop
                return False
            while i > idx:
                self._operations[i-1], self._operations[i] = self._operations[i], self._operations[i-1]
                i -= 1
            sorted[self._operations[i]] = True
        return True

    def move_operation(self, at_idx, to_idx):
        self._operations.insert(to_idx, self._operations.pop(at_idx))

    def reverse(self, offset=0):
        self._operations[offset:] = reversed(self._operations[offset:])

    def __repr__(self):
        return self.name if self.name is not None else _hex_id(self)

    def __str__(self):
        return "graph {name}({inputs}) -> ({outputs})".format(
            name=repr(self),
            inputs=', '.join(repr(input) for input in self.inputs),
            outputs=', '.join(repr(input) for input in self.outputs),
        )

    def print(self, file=None):
        print(f'graph {repr(self)} {{', file=file)

        print(f'\tinputs {{', file=file)
        for tensor in self.inputs:
            print('\t\t' + str(tensor) + ',', file=file)
        print(f'\t}}', file=file)

        print(f'\toutputs {{', file=file)
        for tensor in self.outputs:
            print('\t\t' + str(tensor) + ',', file=file)
        print(f'\t}}', file=file)

        print(f'\tparams {{', file=file)
        for tensor in self.tensors:
            if tensor.producer is None and tensor.data is not None:
                print('\t\t' + str(tensor) + ',', file=file)
        print(f'\t}}', file=file)

        print(f'\toperators {{', file=file)
        for operation in self._operations:
            print('\t\t' + str(operation) + ',', file=file)
        print(f'\t}}', file=file)

        print(f'}}')

    def assert_consistent(self):
        assert len(self.tensors) == len(set(self.tensors))
        assert len(self.operations) == len(set(self.operations))
        for t in self.tensors:
            assert t._graph == self
            assert all(t in consumer.inputs for consumer in t.consumers)
            assert all(t in producer.outputs for producer in t.producers)
            assert all(consumer in self.operations for consumer in t.consumers)
            assert all(producer in self.operations for producer in t.producers)
        for op in self.operations:
            assert op._graph == self
            assert all(op in t.consumers for t in op.inputs)
            assert all(op in t.producers for t in op.outputs)
        for t in self.inputs:
            assert t in self.tensors
        for t in self.outputs:
            assert t in self.tensors


class _ListView(Sequence):

    def __init__(self, lst):
        self._list = lst

    def __len__(self):
        return self._list.__len__()

    def __getitem__(self, item):
        return self._list.__getitem__(item)

    def __iter__(self):
        return self._list.__iter__()

    def __repr__(self):
        return self._list.__repr__()

    def __str__(self):
        return self._list.__str__()

    def __contains__(self, item):
        return self._list.__contains__(item)

    def __reversed__(self):
        return reversed(self._list)


def _hex_id(obj):
    return '@' + hex(id(obj))[2:]



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: tflite

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class Operator(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsOperator(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Operator()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def OperatorBufferHasIdentifier(cls, buf, offset, size_prefixed=False):
        return flatbuffers.util.BufferHasIdentifier(buf, offset, b"\x54\x46\x4C\x33", size_prefixed=size_prefixed)

    # Operator
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Operator
    def OpcodeIndex(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint32Flags, o + self._tab.Pos)
        return 0

    # Operator
    def Inputs(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(flatbuffers.number_types.Int32Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return 0

    # Operator
    def InputsAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Int32Flags, o)
        return 0

    # Operator
    def InputsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Operator
    def InputsIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        return o == 0

    # Operator
    def Outputs(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(flatbuffers.number_types.Int32Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return 0

    # Operator
    def OutputsAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Int32Flags, o)
        return 0

    # Operator
    def OutputsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Operator
    def OutputsIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        return o == 0

    # Operator
    def BuiltinOptionsType(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 0

    # Operator
    def BuiltinOptions(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            from flatbuffers.table import Table
            obj = Table(bytearray(), 0)
            self._tab.Union(obj, o)
            return obj
        return None

    # Operator
    def CustomOptions(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1))
        return 0

    # Operator
    def CustomOptionsAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Operator
    def CustomOptionsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Operator
    def CustomOptionsIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        return o == 0

    # Operator
    def CustomOptionsFormat(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int8Flags, o + self._tab.Pos)
        return 0

    # Operator
    def MutatingVariableInputs(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(flatbuffers.number_types.BoolFlags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1))
        return 0

    # Operator
    def MutatingVariableInputsAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.BoolFlags, o)
        return 0

    # Operator
    def MutatingVariableInputsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Operator
    def MutatingVariableInputsIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        return o == 0

    # Operator
    def Intermediates(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(flatbuffers.number_types.Int32Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return 0

    # Operator
    def IntermediatesAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Int32Flags, o)
        return 0

    # Operator
    def IntermediatesLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Operator
    def IntermediatesIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        return o == 0

def OperatorStart(builder): builder.StartObject(9)
def OperatorAddOpcodeIndex(builder, opcodeIndex): builder.PrependUint32Slot(0, opcodeIndex, 0)
def OperatorAddInputs(builder, inputs): builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(inputs), 0)
def OperatorStartInputsVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def OperatorAddOutputs(builder, outputs): builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(outputs), 0)
def OperatorStartOutputsVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def OperatorAddBuiltinOptionsType(builder, builtinOptionsType): builder.PrependUint8Slot(3, builtinOptionsType, 0)
def OperatorAddBuiltinOptions(builder, builtinOptions): builder.PrependUOffsetTRelativeSlot(4, flatbuffers.number_types.UOffsetTFlags.py_type(builtinOptions), 0)
def OperatorAddCustomOptions(builder, customOptions): builder.PrependUOffsetTRelativeSlot(5, flatbuffers.number_types.UOffsetTFlags.py_type(customOptions), 0)
def OperatorStartCustomOptionsVector(builder, numElems): return builder.StartVector(1, numElems, 1)
def OperatorAddCustomOptionsFormat(builder, customOptionsFormat): builder.PrependInt8Slot(6, customOptionsFormat, 0)
def OperatorAddMutatingVariableInputs(builder, mutatingVariableInputs): builder.PrependUOffsetTRelativeSlot(7, flatbuffers.number_types.UOffsetTFlags.py_type(mutatingVariableInputs), 0)
def OperatorStartMutatingVariableInputsVector(builder, numElems): return builder.StartVector(1, numElems, 1)
def OperatorAddIntermediates(builder, intermediates): builder.PrependUOffsetTRelativeSlot(8, flatbuffers.number_types.UOffsetTFlags.py_type(intermediates), 0)
def OperatorStartIntermediatesVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def OperatorEnd(builder): return builder.EndObject()



# Copyright (c) 2020 The Khronos Group Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from caffe2.python.onnx.backend import Caffe2Backend
from ..onnx.writer import build_model, build_dtype
from ..onnx.reader import _get_value_info
from ...utils.types import as_str
import json
import os


def save_caffe2_model(folder, init_net, predict_net, value_info):
    with open(os.path.join(folder, 'init_net.pb'), 'wb') as file:
        file.write(init_net.SerializeToString())

    with open(os.path.join(folder, 'predict_net.pb'), 'wb') as file:
        file.write(predict_net.SerializeToString())

    with open(os.path.join(folder, 'value_info.json'), 'w') as file:
        json.dump(value_info, file)


def get_value_info(onnx_model):
    initializer_names = {as_str(info.name) for info in onnx_model.graph.initializer}

    value_info = {}
    for info in onnx_model.graph.input:
        name, shape, dtype = _get_value_info(info)
        if name not in initializer_names:
            value_info[name] = (build_dtype(dtype), shape)

    return value_info


class Writer:

    def __init__(self):
        pass

    def __call__(self, graph, folder):
        onnx_model = build_model(graph, ir_version=6, opset_version=9)
        if not onnx_model.graph.name:
            onnx_model.graph.name = 'Graph'

        init_net, predict_net = Caffe2Backend.onnx_graph_to_caffe2_net(onnx_model)
        value_info = get_value_info(onnx_model)

        if not os.path.exists(folder):
            os.mkdir(folder)

        save_caffe2_model(folder, init_net, predict_net, value_info)



# Copyright (c) 2020 The Khronos Group Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..model.utils import replace_chain
from ..model.graph import *
from ..utils.types import from_numpy


class Optimizer:

    def __init__(self, custom_optimizers=None):
        self._custom_optimizers = custom_optimizers or {}

    def __call__(self, graph, only_required=False):
        self._fix_inputs_without_producer(graph)
        replace_chain(graph, ['SpaceToBatchND', {'Conv2D', 'DepthwiseConv2dNative'}, 'BatchToSpaceND'], self._replace_dilated_conv)
        replace_chain(graph, ['Cast'], self._replace_bool_cast)
        for chain, replacer in six.iteritems(self._custom_optimizers):
            replace_chain(graph, chain, replacer)
        if not only_required:
            self._remove_unused_constants(graph)
        return graph

    @staticmethod
    def _fix_inputs_without_producer(graph):
        idx = 0
        for tensor in graph.inputs:
            if tensor.producer is None:
                cnt = len(graph.operations)
                Operation(tensor.graph, type='Placeholder', name=Optimizer._op_name(tensor.name), outputs=tensor,
                          attribs={'shape': tensor.shape, 'dtype': tensor.dtype})
                graph.move_operation(cnt, idx)
                idx += 1
        return idx > 0

    @staticmethod
    def _op_name(tensor_name):
        idx = tensor_name.find(':')
        return tensor_name[:idx] if idx != -1 else tensor_name

    @staticmethod
    def _remove_unused_constants(graph):
        ops = [op for op in graph.operations if op.type == 'Const' and len(op.output.consumers) == 0]
        tensors = [op.output for op in ops]
        graph.remove_operations(ops, unlink=True)
        graph.remove_tensors(tensors)

    @staticmethod
    def _replace_dilated_conv(space_to_batch, conv, batch_to_space):
        if not Optimizer._is_constant(space_to_batch.inputs[1]) or not Optimizer._is_constant(batch_to_space.inputs[1]):
            return False

        block_shape1 = Optimizer._read_constant(space_to_batch.inputs[1])
        block_shape2 = Optimizer._read_constant(batch_to_space.inputs[1])

        if not np.all(block_shape1 == block_shape2):
            return False

        if conv.attribs['padding'] != 'VALID':
            return False

        dilations = from_numpy(block_shape1)

        input = space_to_batch.inputs[0]
        filter = conv.inputs[1]
        output = batch_to_space.outputs[0]

        is_nxc = Optimizer._is_nxc(conv.attribs['data_format'])
        same_padding = Optimizer._is_same_padded(input.shape, output.shape, conv.attribs['strides'], is_nxc)

        if not same_padding:
            return False

        op = conv.copy_with(inputs=(input, filter, *conv.inputs[2:]), outputs=output, attribs=dict(conv.attribs))

        op.attribs['dilations'] = [1] + dilations + [1] if is_nxc else [1, 1] + dilations
        op.attribs['padding'] = 'SAME'
        if '_output_shapes' in op.attribs:
            op.attribs['_output_shapes'] = batch_to_space.attribs['_output_shapes']

    @staticmethod
    def _replace_bool_cast(cast):
        if cast.input.dtype == bool and cast.output.dtype != bool:
            ones = Tensor(cast.graph, name=cast.name + '/ones', dtype=cast.output.dtype, shape=cast.output.shape,
                          data=np.full(fill_value=1, dtype=cast.output.dtype, shape=cast.output.shape))
            zeros = Tensor(cast.graph, name=cast.name + '/zeros', dtype=cast.output.dtype, shape=cast.output.shape,
                           data=np.full(fill_value=0, dtype=cast.output.dtype, shape=cast.output.shape))
            Optimizer._make_constant_producer(ones)
            Optimizer._make_constant_producer(zeros)
            Operation(cast.graph, type='Select', name=cast.name, inputs=(cast.input, ones, zeros), outputs=cast.output,
                      attribs={'T': cast.output.dtype})
        elif cast.input.dtype != bool and cast.output.dtype == bool:
            zeros = Tensor(cast.graph, name=cast.name + '/zeros', dtype=cast.input.dtype, shape=(),
                           data=np.array(0, dtype=cast.input.dtype))
            Optimizer._make_constant_producer(zeros)
            Operation(cast.graph, type='NotEqual', name=cast.name, inputs=(cast.input, zeros), outputs=cast.output,
                      attribs={'T': cast.output.dtype})
        else:
            return False

    @staticmethod
    def _is_constant(tensor):
        return tensor.producer.type == 'Const' if tensor.producer else tensor.data is not None

    @staticmethod
    def _read_constant(tensor):
        return tensor.producer.attribs['value'] if tensor.producer else tensor.data

    @staticmethod
    def _is_nxc(format):
        return format[0] == 'N' and format[-1] == 'C' and len(format) > 2

    @staticmethod
    def _is_same_padded(input, output, stride, is_nxc):
        rank = len(input)
        return all(output[i] == (input[i] + stride[i] - 1) // stride[i]
                   for i in (range(1, rank - 1) if is_nxc else range(2, rank)))

    @staticmethod
    def _make_constant_producer(tensor):
        Operation(tensor.graph, type='Const', name=tensor.name, outputs=tensor,
                  attribs={'dtype': tensor.dtype, 'value': tensor.data})



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: tflite

class BuiltinOptions(object):
    NONE = 0
    Conv2DOptions = 1
    DepthwiseConv2DOptions = 2
    ConcatEmbeddingsOptions = 3
    LSHProjectionOptions = 4
    Pool2DOptions = 5
    SVDFOptions = 6
    RNNOptions = 7
    FullyConnectedOptions = 8
    SoftmaxOptions = 9
    ConcatenationOptions = 10
    AddOptions = 11
    L2NormOptions = 12
    LocalResponseNormalizationOptions = 13
    LSTMOptions = 14
    ResizeBilinearOptions = 15
    CallOptions = 16
    ReshapeOptions = 17
    SkipGramOptions = 18
    SpaceToDepthOptions = 19
    EmbeddingLookupSparseOptions = 20
    MulOptions = 21
    PadOptions = 22
    GatherOptions = 23
    BatchToSpaceNDOptions = 24
    SpaceToBatchNDOptions = 25
    TransposeOptions = 26
    ReducerOptions = 27
    SubOptions = 28
    DivOptions = 29
    SqueezeOptions = 30
    SequenceRNNOptions = 31
    StridedSliceOptions = 32
    ExpOptions = 33
    TopKV2Options = 34
    SplitOptions = 35
    LogSoftmaxOptions = 36
    CastOptions = 37
    DequantizeOptions = 38
    MaximumMinimumOptions = 39
    ArgMaxOptions = 40
    LessOptions = 41
    NegOptions = 42
    PadV2Options = 43
    GreaterOptions = 44
    GreaterEqualOptions = 45
    LessEqualOptions = 46
    SelectOptions = 47
    SliceOptions = 48
    TransposeConvOptions = 49
    SparseToDenseOptions = 50
    TileOptions = 51
    ExpandDimsOptions = 52
    EqualOptions = 53
    NotEqualOptions = 54
    ShapeOptions = 55
    PowOptions = 56
    ArgMinOptions = 57
    FakeQuantOptions = 58
    PackOptions = 59
    LogicalOrOptions = 60
    OneHotOptions = 61
    LogicalAndOptions = 62
    LogicalNotOptions = 63
    UnpackOptions = 64
    FloorDivOptions = 65
    SquareOptions = 66
    ZerosLikeOptions = 67
    FillOptions = 68
    BidirectionalSequenceLSTMOptions = 69
    BidirectionalSequenceRNNOptions = 70
    UnidirectionalSequenceLSTMOptions = 71
    FloorModOptions = 72
    RangeOptions = 73
    ResizeNearestNeighborOptions = 74
    LeakyReluOptions = 75
    SquaredDifferenceOptions = 76
    MirrorPadOptions = 77
    AbsOptions = 78
    SplitVOptions = 79
    UniqueOptions = 80
    ReverseV2Options = 81
    AddNOptions = 82
    GatherNdOptions = 83
    CosOptions = 84
    WhereOptions = 85
    RankOptions = 86
    ReverseSequenceOptions = 87
    MatrixDiagOptions = 88
    QuantizeOptions = 89
    MatrixSetDiagOptions = 90
    HardSwishOptions = 91
    IfOptions = 92
    WhileOptions = 93
    DepthToSpaceOptions = 94
    NonMaxSuppressionV4Options = 95
    NonMaxSuppressionV5Options = 96
    ScatterNdOptions = 97
    SelectV2Options = 98
    DensifyOptions = 99
    SegmentSumOptions = 100
    BatchMatMulOptions = 101




# Copyright (c) 2020 The Khronos Group Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import division, print_function, absolute_import

import torch
import nnef
import os

from .nnef_module import NNEFModule
from .. import Statistics


class Interpreter:

    def __init__(self, model, device=None, decomposed=None, custom_operators=None):
        if isinstance(model, nnef.Graph):
            self._nnef_graph = model
        else:
            self._nnef_graph = nnef.parse_file(os.path.join(model, 'graph.nnef'), lowered=decomposed)
        self._init_input_shapes(self._nnef_graph)

        self._nnef_module = NNEFModule(model=model, custom_operators=custom_operators, decomposed=decomposed)

        if device is None:
            device = 'cuda' if torch.cuda.is_available() else 'cpu'

        self._nnef_module.to(device)
        self._nnef_module.eval()
        self._device = device

    def __call__(self, inputs, output_names=None, collect_statistics=False):
        outputs = {}
        statistics = {} if collect_statistics else None

        def callback(name, tensor):
            if output_names is not None and name in output_names:
                outputs[name] = tensor.detach().cpu().numpy()
            if collect_statistics:
                statistics[name] = self._compute_statistics(tensor)

        if output_names is not None:
            assert all(name in self._nnef_graph.tensors for name in output_names), \
                "could not find tensor(s) named {}".format({name for name in output_names
                                                            if name not in self._nnef_graph.tensors})

        if output_names is not None or collect_statistics:
            self._nnef_module.activation_callback = callback

        torch_inputs = [torch.tensor(input).to(self._device) for input in inputs]
        with torch.no_grad():  # Without this, gradients are calculated even in eval mode
            torch_outputs = self._nnef_module.forward(*torch_inputs)

        self._nnef_module.activation_callback = None

        if output_names is None:
            outputs = {name: torch_tensor.detach().cpu().numpy()
                       for name, torch_tensor in zip(self._nnef_graph.outputs, torch_outputs)}

        return (outputs, statistics) if collect_statistics else outputs

    def input_details(self):
        return [self._nnef_graph.tensors[name] for name in self._nnef_graph.inputs]

    def output_details(self):
        return [self._nnef_graph.tensors[name] for name in self._nnef_graph.outputs]

    def tensor_details(self):
        return self._nnef_graph.tensors.values()

    @staticmethod
    def _compute_statistics(torch_tensor):
        num = torch_tensor.numel()
        if num == 0:
            return Statistics(num=0, min=0.0, max=0.0, sum=0.0, ssum=0.0)
        else:
            return Statistics(
                num=num,
                min=float(torch.min(torch_tensor)),
                max=float(torch.max(torch_tensor)),
                sum=float(torch.sum(torch_tensor)),
                ssum=float(torch.sum(torch_tensor * torch_tensor)),
            )

    @staticmethod
    def _init_input_shapes(graph):
        from nnef.shapes import _set_shape
        for op in graph.operations:
            if op.name == 'external':
                _set_shape(graph, op.outputs['output'], op.attribs['shape'])



# Copyright (c) 2020 The Khronos Group Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import division, print_function, absolute_import

import nnef
import torch
import keyword

from . import nnef_operators
from ...io import nnef as nnef_io
from ...io.nnef.reader import _build_graph


class NNEFModule(torch.nn.Module):

    """
    A torch.nn.Module that interprets the given NNEF model
    """

    def __init__(self,
                 model,  # type: str
                 decomposed=None,  # type: typing.Optional[typing.List[str]]
                 custom_operators=None,  # type: typing.Optional[typing.Dict[str, typing.Callable]]
                 activation_callback=None,  # type: typing.Optional[typing.Callable[[str, torch.Tensor], None]]
                 training_attributes=None,  # type: typing.Optional[typing.Dict[str, typing.Dict[str, typing.Any]]]
                 ):
        # type: (...)->None
        """
            nnef_graph might be modified by this class if training and write_nnef is used
        """
        super(NNEFModule, self).__init__()
        if isinstance(model, nnef.Graph):
            self._nnef_graph = _build_graph(model)
        else:
            reader = nnef_io.Reader(decomposed=decomposed, infer_shapes=False)
            self._nnef_graph = reader(model)

        self._name_inline_constants(self._nnef_graph)

        for nnef_tensor in self._nnef_graph.tensors:
            if self._is_variable(nnef_tensor):
                name = self._registered_name(nnef_tensor.name)
                data = self._dequantize(nnef_tensor.data, nnef_tensor.quant, channel_axis=0) \
                    if nnef_tensor.quant else nnef_tensor.data
                self.register_parameter(name, torch.nn.Parameter(torch.tensor(self.normalize_dtype(data))))
            elif self._is_constant(nnef_tensor):
                name = self._registered_name(nnef_tensor.name)
                data = nnef_tensor.data if not nnef_tensor.producer else \
                    self._as_numpy(nnef_tensor.producer.attribs['value'],
                                   nnef_tensor.producer.attribs['shape'],
                                   nnef_tensor.producer.attribs['dtype']) \

                self.register_buffer(name, torch.tensor(self.normalize_dtype(data)))

        self._operators = {}
        self._operators.update(nnef_operators.Operators)
        if custom_operators:
            self._operators.update(custom_operators)
        self._activation_callback = activation_callback
        self._training_attributes = training_attributes or {}

    def forward(self, *inputs):
        assert len(inputs) == len(self._nnef_graph.inputs)
        activations = {nnef_tensor.name: torch_tensor for torch_tensor, nnef_tensor
                       in zip(inputs, self._nnef_graph.inputs)}

        def get_tensor(name):
            if hasattr(self, self._registered_name(name)):
                return getattr(self, self._registered_name(name))
            else:
                return activations[name]

        def has_tensor(name):
            return hasattr(self, self._registered_name(name)) or name in activations

        for op in self._nnef_graph.operations:
            if op.type == 'external' or op.type == 'variable' or op.type == 'constant':
                output = get_tensor(op.output.name)
                if self._activation_callback:
                    self._activation_callback(op.output.name, output)
            else:
                assert op.type in self._operators, "Unsupported operation: {}".format(op.type)
                func = self._operators[op.type]

                assert all(has_tensor(tensor.name) for tensor in op.inputs),\
                    "could not fetch input tensor(s) {} for operation {}"\
                        .format({tensor.name for tensor in op.inputs}, op.type)

                training_attribs = self._training_attributes.get(op.type, {})
                attribs = {**op.attribs, **training_attribs}
                attribs = {self._escape_keyword(name): value for name, value in six.iteritems(attribs)}

                if 'dtype' in attribs and op.type != 'constant' and op.type != 'cast':
                    del attribs['dtype']

                inputs = [get_tensor(tensor.name) if tensor.name else torch.tensor(tensor.data) for tensor in op.inputs]
                outputs = func(*inputs, **attribs) if isinstance(op.inputs, tuple) else func(inputs, **attribs)

                if not isinstance(outputs, (list, tuple)):
                    outputs = (outputs,)

                for nnef_tensor, output in zip(op.outputs, outputs):
                    if nnef_tensor.quant and not self._is_variable(nnef_tensor):
                        output = self._fake_quantize(output, nnef_tensor.quant, channel_axis=0)

                    activations[nnef_tensor.name] = output
                    if self._activation_callback:
                        self._activation_callback(nnef_tensor.name, output)

                for nnef_tensor in op.inputs:
                    if nnef_tensor.name in activations and op is nnef_tensor.consumers[-1] and \
                            nnef_tensor not in self._nnef_graph.outputs:
                        del activations[nnef_tensor.name]

        return tuple(get_tensor(nnef_tensor.name) for nnef_tensor in self._nnef_graph.outputs)

    def save_nnef(self, path):
        for nnef_tensor in self._nnef_graph.tensors:
            if self._is_variable(nnef_tensor.name):
                torch_tensor = getattr(self, self._registered_name(nnef_tensor.name))
                nnef_tensor.data = torch_tensor.detach().cpu().numpy().astype(nnef_tensor.dtype)

        writer = nnef_io.Writer()
        writer(self._nnef_graph, path)

    @property
    def activation_callback(self):
        return self._activation_callback

    @activation_callback.setter
    def activation_callback(self, callback):
        self._activation_callback = callback

    @staticmethod
    def _is_variable(tensor):
        return tensor.producer and tensor.producer.type == 'variable'

    @staticmethod
    def _is_constant(tensor):
        return not tensor.producer or tensor.producer.type == 'constant'

    @staticmethod
    def _as_numpy(value, shape, dtype):
        if isinstance(value, list):
            if len(value) == 1 and int(np.prod(shape)) != 1:
                return np.full(shape, value[0], dtype=dtype)
            else:
                return np.array(value, dtype=dtype).reshape(shape)
        else:
            return np.full(shape, value, dtype=dtype)

    @staticmethod
    def _escape_keyword(name):
        return name if not keyword.iskeyword(name) else '_' + name + '_'

    @staticmethod
    def _name_inline_constants(graph):
        constants = 0
        for tensor in graph.tensors:
            if not tensor.name:
                assert not tensor.producer
                tensor.name = '$' + str(constants)
                constants += 1

    @staticmethod
    def _registered_name(name):
        return '_nnef_' + name

    @staticmethod
    def normalize_dtype(data):
        dtype = NNEFModule._dtypeRemap.get(data.dtype.type)
        return data.astype(dtype) if dtype is not None else data

    @staticmethod
    def _dequantize(data, quant, channel_axis):
        op_name = quant['op-name']
        rank = len(data.shape)
        if op_name == 'zero_point_linear_quantize':
            return NNEFModule._dequantize_zero_point(data,
                                                     NNEFModule._ensure_rank(quant['zero_point'], rank, channel_axis),
                                                     NNEFModule._ensure_rank(quant['scale'], rank, channel_axis))
        elif op_name == 'min_max_linear_quantize' or op_name == 'linear_quantize':
            return NNEFModule._dequantize_min_max(data,
                                                  NNEFModule._ensure_rank(quant['min'], rank, channel_axis),
                                                  NNEFModule._ensure_rank(quant['max'], rank, channel_axis),
                                                  quant['signed'], quant['symmetric'], quant['bits'])
        else:
            raise ValueError("Quantization operation '{}' not implemented".format(op_name))

    @staticmethod
    def _dequantize_zero_point(data, zero_point, scale):
        return (data - zero_point) * scale

    @staticmethod
    def _dequantize_min_max(data, min, max, signed, symmetric, bits):
        if signed:
            data += 2 ** (bits - 1) - int(symmetric)
        r = 2 ** bits - 1 - int(signed and symmetric)
        return data * ((max - min) / r) + min

    def _fake_quantize(self, tensor, quant, channel_axis):
        op_type = quant['op-name']
        rank = len(tensor.shape)
        attribs = {key: NNEFModule._ensure_rank(value, rank, channel_axis) if isinstance(value, np.ndarray) else value
                   for key, value in six.iteritems(quant) if key != 'op-name'}

        assert op_type in self._operators, "Unsupported quantization operation: {}".format(op_type)
        func = self._operators[op_type]
        return func(tensor, **attribs)

    @staticmethod
    def _ensure_rank(value, rank, offset=0):
        array = np.array(value)
        return np.reshape(array, newshape=(1,) * offset + array.shape + (1,) * (rank - offset - len(array.shape)))

    _dtypeRemap = {
        np.float16: np.float32,
        np.float64: np.float32,
        np.int8: np.int64,
        np.uint8: np.int64,
        np.int16: np.int64,
        np.uint16: np.int64,
        np.int32: np.int64,
        np.uint32: np.int64,
        np.uint64: np.int64,
    }



# Copyright (c) 2020 The Khronos Group Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .reader import Reader
from .writer import Writer
from .composite import replace_composites_with_py_functions, reset_composites
from .utils import set_input_shapes, fold_constant_tensors, retain_reachables_from_outputs, insert_rename_identities
from .utils import import_graph_def, export_graph_def, check_finite, check_variables
from tensorflow.python.framework import graph_util
try:
    import tensorflow.compat.v1 as tf
except ImportError:
    import tensorflow as tf


composite_function = composite.function


def save_default_graph(filename, session, outputs, input_shapes=None, fold_constants=True, collapse_composites=True):
    check_variables(session)

    if not isinstance(outputs, dict):
        outputs = {tensor: tensor.name[:-2] if tensor.name.endswith(':0') else tensor.name.replace(':', '_')
                   for tensor in outputs}

    output_names = list(outputs.values())

    graph_def = export_graph_def(tf.get_default_graph())
    graph_def = insert_rename_identities(graph_def, outputs)
    graph_def = graph_util.convert_variables_to_constants(session, graph_def, output_names)
    graph_def = retain_reachables_from_outputs(graph_def, output_names)

    check_finite(graph_def)

    if input_shapes:
        graph_def = set_input_shapes(graph_def, input_shapes)
    if fold_constants:
        graph_def = fold_constant_tensors(graph_def)
    if collapse_composites:
        graph_def = replace_composites_with_py_functions(graph_def)

    check_finite(graph_def)

    with open(filename, 'wb') as file:
        file.write(graph_def.SerializeToString())


def load_default_graph(filename):
    from .protobuf import GraphDef
    with tf.io.gfile.GFile(filename, "rb") as file:
        graph_def = GraphDef()
        graph_def.ParseFromString(file.read())
        tf.import_graph_def(graph_def, name='')



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: tflite

class BuiltinOperator(object):
    ADD = 0
    AVERAGE_POOL_2D = 1
    CONCATENATION = 2
    CONV_2D = 3
    DEPTHWISE_CONV_2D = 4
    DEPTH_TO_SPACE = 5
    DEQUANTIZE = 6
    EMBEDDING_LOOKUP = 7
    FLOOR = 8
    FULLY_CONNECTED = 9
    HASHTABLE_LOOKUP = 10
    L2_NORMALIZATION = 11
    L2_POOL_2D = 12
    LOCAL_RESPONSE_NORMALIZATION = 13
    LOGISTIC = 14
    LSH_PROJECTION = 15
    LSTM = 16
    MAX_POOL_2D = 17
    MUL = 18
    RELU = 19
    RELU_N1_TO_1 = 20
    RELU6 = 21
    RESHAPE = 22
    RESIZE_BILINEAR = 23
    RNN = 24
    SOFTMAX = 25
    SPACE_TO_DEPTH = 26
    SVDF = 27
    TANH = 28
    CONCAT_EMBEDDINGS = 29
    SKIP_GRAM = 30
    CALL = 31
    CUSTOM = 32
    EMBEDDING_LOOKUP_SPARSE = 33
    PAD = 34
    UNIDIRECTIONAL_SEQUENCE_RNN = 35
    GATHER = 36
    BATCH_TO_SPACE_ND = 37
    SPACE_TO_BATCH_ND = 38
    TRANSPOSE = 39
    MEAN = 40
    SUB = 41
    DIV = 42
    SQUEEZE = 43
    UNIDIRECTIONAL_SEQUENCE_LSTM = 44
    STRIDED_SLICE = 45
    BIDIRECTIONAL_SEQUENCE_RNN = 46
    EXP = 47
    TOPK_V2 = 48
    SPLIT = 49
    LOG_SOFTMAX = 50
    DELEGATE = 51
    BIDIRECTIONAL_SEQUENCE_LSTM = 52
    CAST = 53
    PRELU = 54
    MAXIMUM = 55
    ARG_MAX = 56
    MINIMUM = 57
    LESS = 58
    NEG = 59
    PADV2 = 60
    GREATER = 61
    GREATER_EQUAL = 62
    LESS_EQUAL = 63
    SELECT = 64
    SLICE = 65
    SIN = 66
    TRANSPOSE_CONV = 67
    SPARSE_TO_DENSE = 68
    TILE = 69
    EXPAND_DIMS = 70
    EQUAL = 71
    NOT_EQUAL = 72
    LOG = 73
    SUM = 74
    SQRT = 75
    RSQRT = 76
    SHAPE = 77
    POW = 78
    ARG_MIN = 79
    FAKE_QUANT = 80
    REDUCE_PROD = 81
    REDUCE_MAX = 82
    PACK = 83
    LOGICAL_OR = 84
    ONE_HOT = 85
    LOGICAL_AND = 86
    LOGICAL_NOT = 87
    UNPACK = 88
    REDUCE_MIN = 89
    FLOOR_DIV = 90
    REDUCE_ANY = 91
    SQUARE = 92
    ZEROS_LIKE = 93
    FILL = 94
    FLOOR_MOD = 95
    RANGE = 96
    RESIZE_NEAREST_NEIGHBOR = 97
    LEAKY_RELU = 98
    SQUARED_DIFFERENCE = 99
    MIRROR_PAD = 100
    ABS = 101
    SPLIT_V = 102
    UNIQUE = 103
    CEIL = 104
    REVERSE_V2 = 105
    ADD_N = 106
    GATHER_ND = 107
    COS = 108
    WHERE = 109
    RANK = 110
    ELU = 111
    REVERSE_SEQUENCE = 112
    MATRIX_DIAG = 113
    QUANTIZE = 114
    MATRIX_SET_DIAG = 115
    ROUND = 116
    HARD_SWISH = 117
    IF = 118
    WHILE = 119
    NON_MAX_SUPPRESSION_V4 = 120
    NON_MAX_SUPPRESSION_V5 = 121
    SCATTER_ND = 122
    SELECT_V2 = 123
    DENSIFY = 124
    SEGMENT_SUM = 125
    BATCH_MATMUL = 126




# automatically generated by the FlatBuffers compiler, do not modify

# namespace: tflite

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class Tensor(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsTensor(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Tensor()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def TensorBufferHasIdentifier(cls, buf, offset, size_prefixed=False):
        return flatbuffers.util.BufferHasIdentifier(buf, offset, b"\x54\x46\x4C\x33", size_prefixed=size_prefixed)

    # Tensor
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Tensor
    def Shape(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(flatbuffers.number_types.Int32Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return 0

    # Tensor
    def ShapeAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Int32Flags, o)
        return 0

    # Tensor
    def ShapeLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Tensor
    def ShapeIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        return o == 0

    # Tensor
    def Type(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int8Flags, o + self._tab.Pos)
        return 0

    # Tensor
    def Buffer(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint32Flags, o + self._tab.Pos)
        return 0

    # Tensor
    def Name(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Tensor
    def Quantization(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from .QuantizationParameters import QuantizationParameters
            obj = QuantizationParameters()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Tensor
    def IsVariable(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False

    # Tensor
    def Sparsity(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from .SparsityParameters import SparsityParameters
            obj = SparsityParameters()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Tensor
    def ShapeSignature(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(flatbuffers.number_types.Int32Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return 0

    # Tensor
    def ShapeSignatureAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Int32Flags, o)
        return 0

    # Tensor
    def ShapeSignatureLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Tensor
    def ShapeSignatureIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        return o == 0

def TensorStart(builder): builder.StartObject(8)
def TensorAddShape(builder, shape): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(shape), 0)
def TensorStartShapeVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def TensorAddType(builder, type): builder.PrependInt8Slot(1, type, 0)
def TensorAddBuffer(builder, buffer): builder.PrependUint32Slot(2, buffer, 0)
def TensorAddName(builder, name): builder.PrependUOffsetTRelativeSlot(3, flatbuffers.number_types.UOffsetTFlags.py_type(name), 0)
def TensorAddQuantization(builder, quantization): builder.PrependUOffsetTRelativeSlot(4, flatbuffers.number_types.UOffsetTFlags.py_type(quantization), 0)
def TensorAddIsVariable(builder, isVariable): builder.PrependBoolSlot(5, isVariable, 0)
def TensorAddSparsity(builder, sparsity): builder.PrependUOffsetTRelativeSlot(6, flatbuffers.number_types.UOffsetTFlags.py_type(sparsity), 0)
def TensorAddShapeSignature(builder, shapeSignature): builder.PrependUOffsetTRelativeSlot(7, flatbuffers.number_types.UOffsetTFlags.py_type(shapeSignature), 0)
def TensorStartShapeSignatureVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def TensorEnd(builder): return builder.EndObject()



# Copyright (c) 2020 The Khronos Group Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math


class Statistics:

    def __init__(self, num, min, max, sum, ssum):
        self.num = num
        self.min = min
        self.max = max
        self.sum = sum
        self.ssum = ssum

    def __add__(self, other):
        self.num += other.num
        self.min = min(self.min, other.min)
        self.max = max(self.max, other.max)
        self.sum += other.sum
        self.ssum += other.ssum
        return self

    def mean(self):
        return self.sum / self.num if self.num != 0 else 0.0

    def variance(self, unbiased=True):
        if self.num <= 1:
            return 0.0

        count = self.num - 1 if unbiased else self.num
        return self.ssum / count - self.sum * self.sum / (self.num * count)

    def std(self, unbiased=True):
        return math.sqrt(max(self.variance(unbiased), 0))



# Copyright (c) 2020 The Khronos Group Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import onnx
from onnx.shape_inference import infer_shapes
import caffe2.python.onnx.frontend
from caffe2.proto import caffe2_pb2
from ..onnx.reader import onnx_model_to_graph
from ...utils.types import as_str
from google.protobuf import text_format
from collections.abc import Sequence
from caffe2.python.workspace import GlobalInit
import json
import sys
import os

try:
    import caffe
except ImportError:
    from . import caffe
    sys.modules['caffe'] = caffe

from caffe2.python.caffe_translator import TranslateModel, TranslatorRegistry, ConvertTensorProtosToInitNet, \
    AddArgument, BaseTranslate


@TranslatorRegistry.Register("ArgMax")
def TranslateArgmax(layer, pretrained_blobs, is_test, **kwargs):
    param = layer.argmax_param
    if param.top_k != 1:
        raise ValueError("Unsupported attribute value 'top_k = {}' in ArgMax layer".format(param.top_k))
    if param.out_max_val:
        raise ValueError("Conversion of ArgMax layer with 'out_max_val = True' is not supported")

    if param.HasField("axis"):
        caffe_op = BaseTranslate(layer, "ArgMax")
        AddArgument(caffe_op, "keepdims", True)
        AddArgument(caffe_op, "axis", param.axis)
        return caffe_op, []
    else:
        flatten_op = caffe2_pb2.OperatorDef()
        flatten_op.type = "Flatten"
        flatten_op.input.extend(layer.bottom)
        flatten_op.output.append(layer.bottom[0] + "_flattened")

        argmax_op = caffe2_pb2.OperatorDef()
        argmax_op.type = "ArgMax"
        argmax_op.input.append(layer.bottom[0] + "_flattened")
        argmax_op.output.extend(layer.top)
        AddArgument(argmax_op, "keepdims", True)
        AddArgument(argmax_op, "axis", 1)

        return [flatten_op, argmax_op], []


@TranslatorRegistry.Register("ReLU")
def TranslateRelu(layer, pretrained_blobs, is_test, **kwargs):
    param = layer.relu_param
    if param.HasField("negative_slope") and param.negative_slope:
        relu_op = BaseTranslate(layer, "LeakyRelu")
        AddArgument(relu_op, "alpha", param.negative_slope)
        return relu_op, []
    else:
        return BaseTranslate(layer, "Relu"), []


def _HookedTranslateLayer(layer, pretrained_blobs, is_test, **kwargs):
    _pre_translate(layer, pretrained_blobs)
    ops, params = _TranslateLayer(layer, pretrained_blobs, is_test, **kwargs)
    _post_translate(layer, pretrained_blobs, ops, params)
    return ops, params


_TranslateLayer = TranslatorRegistry.TranslateLayer
TranslatorRegistry.TranslateLayer = _HookedTranslateLayer


GlobalInit(['caffe2', '--caffe2_log_level=2'])


_UnrecognizedAttribs = {'ws_nbytes_limit'}


def _caffe_to_caffe2(prototxt, caffemodel):
    if prototxt.layer[0].type == 'Input':
        input_names = list(prototxt.layer[0].top)
        input_shapes = list(item.dim for item in prototxt.layer[0].input_param.shape)
    else:
        input_names = list(item for item in prototxt.input)
        input_shapes = list(item.dim for item in prototxt.input_shape)
        if len(input_shapes) == 0:
            input_dims = prototxt.input_dim
            assert len(input_dims) == 4 * len(input_names)
            input_shapes = [None] * len(input_names)
            for i in range(len(input_names)):
                input_shapes[i] = input_dims[4 * i: 4 * (i+1)]

    predict_net, params = TranslateModel(prototxt, caffemodel, is_test=True, remove_legacy_pad=False, input_dims=[])

    # Assume there is one input and one output
    external_input = predict_net.op[0].input[0]
    external_output = predict_net.op[-1].output[0]

    predict_net.external_input.extend([external_input])
    predict_net.external_input.extend([param.name for param in params.protos])
    predict_net.external_output.extend([external_output])
    init_net = ConvertTensorProtosToInitNet(params, external_input)

    value_info = {name: (onnx.TensorProto.FLOAT, shape) for name, shape in zip(input_names, input_shapes)}

    return predict_net, init_net, value_info


_DeconvGroups = {}


def _pre_translate(layer, blobs):
    if layer.type == 'Convolution':
        _fix_conv_pool_param(layer.convolution_param)
    elif layer.type == 'Deconvolution':
        _fix_conv_pool_param(layer.convolution_param)
        if layer.convolution_param.group != 1:
            _DeconvGroups[layer.name] = layer.convolution_param.group
            layer.convolution_param.group = 0   # to trick the conversion check
    elif layer.type == 'Pooling':
        _fix_conv_pool_param(layer.pooling_param)
    elif layer.type == 'Eltwise':
        _fix_eltwise_param(layer.eltwise_param)
    elif layer.type == 'BatchNorm':
        _fix_batch_norm_param(layer.batch_norm_param, blobs)


def _post_translate(layer, blobs, ops, params):
    if layer.type == 'Deconvolution' and layer.convolution_param.group == 0:
        AddArgument(ops[0], 'group', _DeconvGroups[layer.name])


def _fix_conv_pool_param(param):
    if isinstance(param.kernel_size, Sequence) and len(param.kernel_size) == 2:
        param.kernel_h = param.kernel_size[0]
        param.kernel_w = param.kernel_size[1]
        del param.kernel_size[1]
        del param.kernel_size[0]
    if isinstance(param.stride, Sequence) and len(param.stride) == 2:
        param.stride_h = param.stride[0]
        param.stride_w = param.stride[1]
        del param.stride[1]
        del param.stride[0]
    if isinstance(param.pad, Sequence) and len(param.pad) == 2:
        param.pad_h = param.pad[0]
        param.pad_w = param.pad[1]
        del param.pad[1]
        del param.pad[0]


def _fix_eltwise_param(param):
    if len(param.coeff) > 0 and all(c == 1 for c in param.coeff):
        for i in reversed(range(len(param.coeff))):
            del param.coeff[i]


def _fix_batch_norm_param(param, blobs):
    if len(blobs) > 2:
        if blobs[2].data[0] == 0:
            blobs[2].data[0] = 1


def _caffe2_net_to_onnx_model(predict_net, init_net, value_info):
    graph = caffe2.python.onnx.frontend.caffe2_net_to_onnx_graph(predict_net, init_net, value_info)
    if not graph.name:
        graph.name = 'Graph'

    opset_id = onnx.OperatorSetIdProto()
    opset_id.domain = ''
    opset_id.version = 11
    model = onnx.helper.make_model(graph, opset_imports=[opset_id])
    onnx.checker.check_model(model)
    return model


def _remove_unrecognized_attributes(net_def):
    for op_def in net_def.op:
        for idx in reversed(range(len(op_def.arg))):
            name = as_str(op_def.arg[idx].name)
            if name in _UnrecognizedAttribs:
                del op_def.arg[idx]


def load_caffe_model(path):
    base, ext = os.path.splitext(path)
    assert ext == '.prototxt'

    with open(path) as file:
        prototxt = caffe.proto.caffe_pb2.NetParameter()
        text_format.Merge(file.read(), prototxt)
    with open(base + '.caffemodel', 'rb') as file:
        caffemodel = caffe.proto.caffe_pb2.NetParameter()
        caffemodel.ParseFromString(file.read())

    return prototxt, caffemodel


def load_caffe_model_as_onnx(path):
    prototxt, caffemodel = load_caffe_model(path)
    predict_net, init_net, value_info = _caffe_to_caffe2(prototxt, caffemodel)
    _remove_unrecognized_attributes(predict_net)
    return _caffe2_net_to_onnx_model(predict_net, init_net, value_info)


def load_caffe2_model(folder):
    predict_net = caffe2_pb2.NetDef()
    with open(os.path.join(folder, 'predict_net.pb'), 'rb') as file:
        predict_net.ParseFromString(file.read())

    init_net = caffe2_pb2.NetDef()
    with open(os.path.join(folder, 'init_net.pb'), 'rb') as file:
        init_net.ParseFromString(file.read())

    with open(os.path.join(folder, 'value_info.json')) as file:
        value_info = json.load(file)

    return predict_net, init_net, value_info


def load_caffe2_model_as_onnx(folder):
    predict_net, init_net, value_info = load_caffe2_model(folder)
    _remove_unrecognized_attributes(predict_net)
    return _caffe2_net_to_onnx_model(predict_net, init_net, value_info)


class Reader:

    def __init__(self, legacy=False):
        self._legacy = legacy

    def __call__(self, path):
        onnx_model = load_caffe_model_as_onnx(path) if self._legacy else load_caffe2_model_as_onnx(path)
        onnx.checker.check_model(onnx_model)
        onnx_model = infer_shapes(onnx_model)
        return onnx_model_to_graph(onnx_model)



# Copyright (c) 2020 The Khronos Group Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import division, print_function, absolute_import

from ...model import *
from ...utils.types import as_str
from onnx.shape_inference import infer_shapes
import numpy as np
import onnx
import sys


_is_little_endian_system = (sys.byteorder == 'little')


_DtypeToNumpy = {
    'UNDEFINED': None,
    'FLOAT': np.float32,
    'UINT8': np.uint8,
    'INT8': np.int8,
    'UINT16': np.uint16,
    'INT16': np.int16,
    'INT32': np.int32,
    'INT64': np.int64,
    'STRING': np.str_,
    'BOOL': np.bool_,
    'FLOAT16': np.float16,
    'DOUBLE': np.float64,
    'UINT32': np.uint32,
    'UINT64': np.uint64,
    'COMPLEX64': np.complex64,
    'COMPLEX128': np.complex128,
}


def _get_shape(tensor_shape_proto):
    return ([int(dim.dim_value) if dim.HasField('dim_value') else None for dim in tensor_shape_proto.dim]
            if tensor_shape_proto is not None else None)


def _get_dtype(dtype_int):
    return _DtypeToNumpy[onnx.TensorProto.DataType.Name(dtype_int)]


def _get_field(proto, name, default=None):
    return getattr(proto, name) if proto.HasField(name) else default


def _get_value_info(value_info_proto):
    name = as_str(value_info_proto.name)
    shape = _get_shape(_get_field(value_info_proto.type.tensor_type, 'shape'))
    dtype = _get_dtype(value_info_proto.type.tensor_type.elem_type)
    return name, shape, dtype


def _get_tensor(tensor_proto):
    assert not tensor_proto.HasField('segment'), 'TensorProto.segment is not supported'

    name = as_str(tensor_proto.name)
    shape = [int(dim) for dim in tensor_proto.dims]
    dtype = _get_dtype(tensor_proto.data_type)
    assert dtype is not None

    if tensor_proto.HasField('raw_data'):
        assert dtype != np.str_

        data = np.frombuffer(tensor_proto.raw_data, dtype)
        if not _is_little_endian_system:
            data = data.byteswap()
    else:
        if dtype == np.float32:
            data = np.array(tensor_proto.float_data, dtype)
        elif dtype == np.float64:
            data = np.array(tensor_proto.double_data, dtype)
        elif dtype == np.int64:
            data = np.array(tensor_proto.int64_data, dtype)
        elif dtype == np.str_:
            data = np.array(as_str(tensor_proto.string_data))
        elif dtype == np.float16:
            data = np.array(tensor_proto.int32_data, np.uint16).view(np.float16)
        elif dtype == np.complex64:
            data = np.array(tensor_proto.float_data, np.float32)
            data = data[0::2] + data[1::2] * 1j
        elif dtype == np.complex128:
            data = np.array(tensor_proto.double_data, np.float64)
            data = data[0::2] + data[1::2] * 1j
        elif dtype in [np.int8, np.uint8, np.int16, np.uint16, np.int32, np.bool_]:
            data = np.array(tensor_proto.int32_data, dtype)
        elif dtype in [np.uint32, np.uint64]:
            data = np.array(tensor_proto.uint64_data, dtype)
        else:
            assert False

    data = data.reshape(shape)
    return name, shape, dtype, data


def _get_tensor_data(tensor_proto):
    name, shape, dtype, data = _get_tensor(tensor_proto)
    return data


def _get_tensors(graph_proto, graph, tensors_by_name):
    for value_info in graph_proto.input:
        name, shape, dtype = _get_value_info(value_info)
        tensors_by_name[name] = Tensor(graph=graph, name=name, shape=shape, dtype=dtype)
    for value_info in graph_proto.output:
        name, shape, dtype = _get_value_info(value_info)
        tensors_by_name[name] = Tensor(graph=graph, name=name, shape=shape, dtype=dtype)
    for value_info in graph_proto.value_info:
        name, shape, dtype = _get_value_info(value_info)
        tensors_by_name[name] = Tensor(graph=graph, name=name, shape=shape, dtype=dtype)
    for tensor_proto in graph_proto.initializer:
        name, shape, dtype, data = _get_tensor(tensor_proto)
        tensors_by_name[name] = Tensor(graph=graph, name=name, shape=shape, dtype=dtype, data=data)

    for node in graph_proto.node:
        for tensor_name in node.output:
            tensor_name = as_str(tensor_name)
            if tensor_name not in tensors_by_name:
                if len(tensor_name) == 0:
                    tensors_by_name[tensor_name] = Tensor(graph, name='', shape=(), dtype=np.float32,
                                                          data=np.zeros(shape=(), dtype=np.float32))
                else:
                    tensors_by_name[tensor_name] = Tensor(graph, name=tensor_name)

    for node in graph_proto.node:
        for attribute in node.attribute:
            if attribute.HasField('g'):
                _get_tensors(attribute.g, graph, tensors_by_name)
            if attribute.graphs:
                for g in attribute.graphs:
                    _get_tensors(g, graph, tensors_by_name)


def _get_node(node_proto, graph, tensors_by_name):
    inputs = [as_str(input) for input in node_proto.input]
    outputs = [as_str(output) for output in node_proto.output]
    name = as_str(_get_field(node_proto, 'name'))
    domain = as_str(_get_field(node_proto, 'domain'))
    op_type = as_str(node_proto.op_type)
    attributes = {}
    for attribute in node_proto.attribute:
        key, value = _get_attribute(attribute, graph, tensors_by_name)
        attributes[key] = value
    return inputs, outputs, name, domain, op_type, attributes


def _get_attribute(attribute_proto, graph, tensors_by_name):
    assert not attribute_proto.HasField('ref_attr_name')

    name = as_str(attribute_proto.name)

    if attribute_proto.HasField('f'):
        value = float(attribute_proto.f)
    elif attribute_proto.HasField('i'):
        value = int(attribute_proto.i)
    elif attribute_proto.HasField('s'):
        value = as_str(attribute_proto.s)
    elif attribute_proto.HasField('t'):
        value = _get_tensor_data(attribute_proto.t)
    elif attribute_proto.HasField('g'):
        g = attribute_proto.g
        value = _get_block(g, Graph(name=as_str(_get_field(g, 'name'))), tensors_by_name)
    elif attribute_proto.floats:
        value = [float(f) for f in attribute_proto.floats]
    elif attribute_proto.ints:
        value = [int(i) for i in attribute_proto.ints]
    elif attribute_proto.strings:
        value = [as_str(s) for s in attribute_proto.strings]
    elif attribute_proto.tensors:
        value = [_get_tensor_data(t) for t in attribute_proto.tensors]
    elif attribute_proto.graphs:
        value = [_get_block(g, Graph(name=as_str(_get_field(g, 'name'))), tensors_by_name)
                 for g in attribute_proto.graphs]
    else:
        value = []

    return name, value


def _get_block(graph_proto, graph, tensors_by_name):
    initializer_names = {as_str(value_info.name) for value_info in graph_proto.initializer}
    input_names = [as_str(value_info.name) for value_info in graph_proto.input
                   if as_str(value_info.name) not in initializer_names]
    output_names = [as_str(value_info.name) for value_info in graph_proto.output]
    graph.inputs = [tensors_by_name[name] for name in input_names]
    graph.outputs = [tensors_by_name[name] for name in output_names]

    for value_info in graph_proto.input:
        name, shape, dtype = _get_value_info(value_info)
        tensor = tensors_by_name[name]
        tensor.shape, tensor.dtype = shape, dtype
    for value_info in graph_proto.output:
        name, shape, dtype = _get_value_info(value_info)
        tensor = tensors_by_name[name]
        tensor.shape, tensor.dtype = shape, dtype
    for value_info in graph_proto.value_info:
        name, shape, dtype = _get_value_info(value_info)
        tensor = tensors_by_name[name]
        tensor.shape, tensor.dtype = shape, dtype
    for tensor_proto in graph_proto.initializer:
        name, shape, dtype, data = _get_tensor(tensor_proto)
        tensor = tensors_by_name[name]
        tensor.shape, tensor.dtype, tensor.data = shape, dtype, data

    for annotation in graph_proto.quantization_annotation:
        tensor = tensors_by_name[annotation.tensor_name]
        tensor.quant = {item.key: tensors_by_name[item.value].data for item in annotation.quant_parameter_tensor_names}

    for node in graph_proto.node:
        inputs, outputs, name, domain, op_type, attributes = _get_node(node, graph, tensors_by_name)

        Operation(
            graph=graph,
            type=op_type,
            name=name,
            inputs=tuple(tensors_by_name[input] for input in inputs),
            outputs=tuple(tensors_by_name[output] for output in outputs),
            attribs=attributes)
    return graph


def _set_input_shapes(graph_proto, input_shapes):
    for value_info in graph_proto.input:
        name, shape, dtype = _get_value_info(value_info)
        input_shape = input_shapes.get(name)
        if input_shape is not None:
            assert len(input_shape) == len(shape) and all(s is None or z == s for s, z in zip(shape, input_shape))
            for i, s in enumerate(input_shape):
                value_info.type.tensor_type.shape.dim[i].dim_value = s


# This is for working around a bug in ONNX IR, see https://github.com/onnx/onnx/issues/2903
def _add_value_info_for_constants(model: onnx.ModelProto):
    """
    Currently onnx.shape_inference doesn't use the shape of initializers, so add
    that info explicitly as ValueInfoProtos.
    Mutates the model.
    Args:
        model: The ModelProto to update.
    """
    # All (top-level) constants will have ValueInfos before IRv4 as they are all inputs
    if model.ir_version < 4:
        return

    def add_const_value_infos_to_graph(graph: onnx.GraphProto):
        inputs = {i.name for i in graph.input}
        existing_info = {vi.name: vi for vi in graph.value_info}
        for init in graph.initializer:
            # Check it really is a constant, not an input
            if init.name in inputs:
                continue

            # The details we want to add
            elem_type = init.data_type
            shape = init.dims

            # Get existing or create new value info for this constant
            vi = existing_info.get(init.name)
            if vi is None:
                vi = graph.value_info.add()
                vi.name = init.name

            # Even though it would be weird, we will not overwrite info even if it doesn't match
            tt = vi.type.tensor_type
            if tt.elem_type == onnx.TensorProto.UNDEFINED:
                tt.elem_type = elem_type
            if not tt.HasField("shape"):
                # Ensure we set an empty list if the const is scalar (zero dims)
                tt.shape.dim.extend([])
                for dim in shape:
                    tt.shape.dim.add().dim_value = dim

        # Handle subgraphs
        for node in graph.node:
            for attr in node.attribute:
                # Ref attrs refer to other attrs, so we don't need to do anything
                if attr.ref_attr_name != "":
                    continue

                if attr.type == onnx.AttributeProto.GRAPH:
                    add_const_value_infos_to_graph(attr.g)
                if attr.type == onnx.AttributeProto.GRAPHS:
                    for g in attr.graphs:
                        add_const_value_infos_to_graph(g)

    return add_const_value_infos_to_graph(model.graph)


def onnx_model_to_graph(onnx_model):
    graph = Graph(name=as_str(_get_field(onnx_model.graph, 'name')))

    tensors_by_name = {'': Tensor(graph, name='', shape=(), dtype=np.float32, data=np.zeros(shape=(), dtype=np.float32))}

    _get_tensors(onnx_model.graph, graph, tensors_by_name)
    _get_block(onnx_model.graph, graph, tensors_by_name)

    return graph


def read_tensor(filename):
    with open(filename, 'rb') as file:
        return _get_tensor_data(onnx.load_tensor(file))


class Reader(object):

    def __init__(self, simplify=False, optimize=None):
        self._simplify = simplify
        self._optimize = optimize or simplify

    def __call__(self, filename, input_shapes=None):
        model_proto = onnx.load_model(filename)
        _add_value_info_for_constants(model_proto)

        if self._simplify:
            from onnxsim import simplify
            model_proto, _ = simplify(model_proto, overwrite_input_shapes=input_shapes, perform_optimization=self._optimize)
        if input_shapes:
            _set_input_shapes(model_proto.graph, input_shapes)

        model_proto = infer_shapes(model_proto)

        return onnx_model_to_graph(model_proto)
